#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Event.py
#	
#	This module is the non-gui class the represents events. Events
#	represent a piece of audio such as a file or something that was recorded.
#	This module handles loading and saving events from xml, the gstreamer
#	bits event properties as well as any event specific functionality like;
#	audio fades, splits, joins, trims, waveform levels, etc.
#
#-------------------------------------------------------------------------------

import xml.dom.minidom as xml
import os, sys, os.path
import pygst
pygst.require("0.10")
import gst, gobject
import Utils, LevelsList
import UndoSystem, IncrementalSave
import Globals
import gettext
import urllib
import PlatformUtils

from elements.singledecodebin import SingleDecodeBin
_ = gettext.gettext

#=========================================================================

class Event(gobject.GObject):
	"""
	This class handles maintaing the information for a single audio 
	event, normally, a fragment of a recorded file.

	Signals:
		"waveform" -- The waveform date for this event has changed.
		"position" -- The starting position of this event has changed.
		"length" -- The length of this event has changed.
		"corrupt" -- The audio file for this event is not playable. Two strings with detailed information are sent.
		"loading" -- Loading has started or completed.

	"""
	__gsignals__ = {
		"waveform" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"position" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"length" 		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"corrupt" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING,)),
		"loading" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"selected" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () )
	}

	""" The level sample interval in seconds """
	LEVEL_INTERVAL = 0.1
	LEVELS_FILE_EXTENSION = ".leveldata"
	
	#_____________________________________________________________________
	
	def __init__(self, instrument, file=None, id=None, filelabel=None):
		"""
		Creates a new instance of Event.
		
		Parameters:
			instrument -- Instrument associated with this Event.
			file -- the file this Event should play.
			id -- unique ID for this Event. If it's taken, a new one will be generated.
			filelabel -- label to print in error messages.
						It can be different	from the file parameter.
		"""
		gobject.GObject.__init__(self)
		
		self.id = instrument.project.GenerateUniqueID(id)  #check is id is already taken, then set it.
		self.start = 0.0			# Time in seconds at which the event begins
		self.duration = 0.0			# Duration in seconds of the event
		# The file this event should play (without escaped characters)
		# If you need characters escaped, please do self.file.replace(" ", "\ ") 
		# but **do not** assign it to this variable.
		self.file = file
		if self.file and os.path.isabs(self.file) and \
		      PlatformUtils.samefile(instrument.project.audio_path, os.path.dirname(self.file)):
			# If the file is in the audio dir, just include the filename, not the absolute path
			Globals.debug("Event() given absolute file, should be relative:", self.file)
			self.file = os.path.basename(self.file)
		
		# levels_file is a filename only, no directory information for levels here.
		basename = os.path.basename(self.file or "Unknown")
		self.levels_file = "%s_%d%s" % (basename, self.id, self.LEVELS_FILE_EXTENSION)
		
		# the label is the filename to print in error messages
		# if it differs from the real filename (i.e its been copied into the project)
		if filelabel != None:
			self.filelabel = filelabel
		else:
			self.filelabel = file

		self.isSelected = False		# True if the event is currently selected
		self.name = "New Event"		# Name of this event
		
		self.selection  = [0, 0]	# List start and end of selection (for fades, etc) measured in seconds 
		self.levels_list = LevelsList.LevelsList()	# LevelsList class containing array of audio levels to be drawn for this event
		
		self.instrument = instrument	# The parent instrument
		self.gnlsrc = None 			# The gstreamer gnlsource object.
		self.single_decode_bin = None		# The gstreamer file decoder element.
		
		self.offset = 0.0			# Offset through the file in seconds
		self.isLoading = False		# True if the event is currently loading level data
		self.isDownloading = False	# True if the event is currently loading from a remote source.
		self.isRecording = False		# True if the event is currently loading level data from a live recording
		self.loadingLength = 0 		# The length of the file in seconds as its being rendered
		self.lastEnd = 0 			# The last length of the loading file - used to minimise redraws
		self.loadingPipeline = None	# The Gstreamer pipeline used to load the waveform
		self.bus = None			# The bus to monitor messages on the loadingPipeline
		
		self.CreateFilesource()

		# a private dictionary containing the audio fade point times as keys
		# and the volume for that point between 0 and 1 as the values.
		# this is private, so if someone else wants a list of audio fade points
		# they must use the sorted list below.
		self.__fadePointsDict = {}
		# A list of control points for the audio fades
		# where each tuple is (<time in seconds>, <volume between 0 and 1>)
		# The list *must* be ordered by time-in-seconds, so when you update it from
		# the dictionary using dict.items(), be sure to sort it again.
		self.audioFadePoints = []
		#Just like self.levels_list except with all the levels scaled according to the
		#points in self.audioFadePoints.
		self.fadeLevels = LevelsList.LevelsList()

	#_____________________________________________________________________
	
	def GetFilename(self):
		return os.path.basename(self.file)
	
	#_____________________________________________________________________
	
	def GetAbsFile(self):
		if os.path.isabs(self.file):
			return self.file
		else:
			return os.path.join(self.instrument.project.audio_path, self.file)
	
	#_____________________________________________________________________
	
	def GetAbsLevelsFile(self):
		return os.path.join(self.instrument.project.levels_path, self.levels_file)
	
	#_____________________________________________________________________
	
	def CreateFilesource(self):	
		"""
		Creates a new GStreamer file source with an unique id.
		It then calls SetProperties() to populate the new object's
		properties.
		"""
		Globals.debug("create file source")
		if not self.gnlsrc:
			self.gnlsrc = gst.element_factory_make("gnlsource", "Event_%d"%self.id)
		if not self.gnlsrc in list(self.instrument.composition.elements()):
			self.instrument.composition.add(self.gnlsrc)
		
		self.SetProperties()

	#_____________________________________________________________________

	def DestroyFilesource(self):
		"""
		Removes the Gstreamer file source from the instrument's composition.
		"""
		if self.gnlsrc in list(self.instrument.composition.elements()):
			self.instrument.composition.remove(self.gnlsrc)
		
	#_____________________________________________________________________
		
	def SetProperties(self):
		"""
		Sets basic Event properties like location, start, duration, etc.
		"""
		if self.file:
			if self.single_decode_bin:
				self.gnlsrc.remove(self.single_decode_bin)
				self.single_decode_bin.set_state(gst.STATE_NULL)

			Globals.debug("creating SingleDecodeBin")
			caps = gst.caps_from_string("audio/x-raw-int;audio/x-raw-float")
			f = PlatformUtils.pathname2url(self.GetAbsFile())
			Globals.debug("file uri is:", f)
			self.single_decode_bin = SingleDecodeBin(caps=caps, uri=f)
			self.gnlsrc.add(self.single_decode_bin)
			Globals.debug("setting event properties:")
			propsDict = {
					"caps" : caps,
					"start" : long(self.start * gst.SECOND),
					"duration" : long(self.duration * gst.SECOND),
					"media-start" : long(self.offset * gst.SECOND),
					"media-duration" : long(self.duration * gst.SECOND),
					"priority" : 2
					}
					
			for prop, value in propsDict.iteritems():
				self.gnlsrc.set_property(prop, value)
				Globals.debug("\t", prop, "=", value)

	#_____________________________________________________________________

	def StoreToXML(self, doc, parent, graveyard=False):
		"""
		Converts this Event into an XML representation suitable for saving to a file.

		Parameters:
			doc -- the XML document object the Event will be saved to.
			parent -- the parent node that the serialized Event should
						be added to.
			graveyard -- True if this Event is on the graveyard stack,
						and should be serialized as a dead Event.
		"""
		if graveyard:
			ev = doc.createElement("DeadEvent")
		else:
			ev = doc.createElement("Event")
		parent.appendChild(ev)
		ev.setAttribute("id", str(self.id))
		
		params = doc.createElement("Parameters")
		ev.appendChild(params)
		
		items = ["start", "duration", "isSelected", 
				  "name", "offset", "file", "filelabel", "levels_file",
				  "isLoading", "isRecording"
				]
				
		#Since we are saving the path to the project file, don't delete it on exit
		if self.GetAbsFile() in self.instrument.project.deleteOnCloseAudioFiles:
			self.instrument.project.deleteOnCloseAudioFiles.remove(self.GetAbsFile())
		
		Utils.StoreParametersToXML(self, doc, params, items)		
		
		xmlPoints = doc.createElement("FadePoints")
		ev.appendChild(xmlPoints)
		Utils.StoreDictionaryToXML(doc, xmlPoints, self.__fadePointsDict, "FadePoint")
		
		if self.levels_list:
			self.levels_list.tofile(self.GetAbsLevelsFile())
		if self.GetAbsLevelsFile() in self.instrument.project.deleteOnCloseAudioFiles:
			self.instrument.project.deleteOnCloseAudioFiles.remove(self.GetAbsLevelsFile())
		
	#_____________________________________________________________________
		
	def __repr__(self):
		"""
		Creates a representation string of the Event.

		Returns:
			a string representing the name, id, start and duration
			for this Event.
		"""
		return "Event: '%s' (%d) : %d -> %d" % (self.name, self.id, self.start, self.duration)
		
	#_____________________________________________________________________
	
	def __cmp__(self, object):
		"""
		Compares two Events for equality.
		
		Returns:
			True -- the Events are equal.
			False -- the Events are different.
		"""
		if type(object) != type(self):
			return -1
		else:
			# If cmp(self.start, object.start) is not zero, return it.
			# If it is zero, return cmp(seld.id, object.id).
			return cmp(self.start, object.start) or cmp(self.id, object.id)
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("Move", "temp")
	def Move(self, to, frm=None):
		"""
		Moves this Event in time.

		Parameters:
			to -- the time the Event's moving to.
			frm -- the time the Event's moving from.
		"""
		if frm is None:
			self.temp = self.start
		else:
			self.temp = frm
		self.start = to
		self.SetProperties()
		self.emit("position")
	
	#_____________________________________________________________________
	
	def _Compat09_Move(self, frm, to, _undoAction_):
		"""
		Moves this Event in time.
		
		Considerations:
			A compatibility method for undo functions from 
			0.1, 0.2 and 0.9 project files. It should not be called
			explicitly by anyone.
		
		Parameters:
			frm -- the time the Event's moving from.
			to -- the time the Event's moving to.
		"""
		self.Move(to, frm, _undoAction_=_undoAction_)
	
	#_____________________________________________________________________
	
	def _Compat02_Split(self, split_point, id=-1):
		"""
		Function kept for compatibility with 0.2 project files.
		Parameters are the same as SplitEvent().
		"""
		self.SplitEvent(split_point)
	
	#_____________________________________________________________________
	
	def _Compat02_Join(self, joinEventID):
		"""
		Function kept for compatibility with 0.2 project files.
		Parameters are the same as JoinEvent().
		"""
		self.JoinEvent(joinEventID)
	
	#_____________________________________________________________________
	
	def CopySelection(self, eventID=-1):
		"""
		Only for use with a 2-point selection.
		Essentially performs a 'fake split' and returns a new event
		which would be the result of splitting an event at the 2 points.
		
		This is used when the user shift-drags an event to create a selection,
		then chooses 'copy' from the context menu. The new event can be placed
		wherever the user wishes by right-clicking and choosing 'paste'.
		"""
		if eventID >= 0:
			e = [x for x in self.instrument.graveyard if x.id == eventID][0]
			self.instrument.graveyard.remove(e)
		else:
			e = Event(self.instrument, self.file)
		e.name = self.name
		
		dur = self.selection[1] - self.selection[0]
		
		e.start = self.start + self.selection[0]
		e.offset = self.selection[0] #+self.offset
		e.duration = dur
		
		dictLeft = {}
		dictRight = {}
		for key, value in self.__fadePointsDict.iteritems():
			if key < self.selection[0]:
				dictLeft[key] = value
			if key > self.selection[0]:
				dictRight[key - self.selection[0]] = value
		#in case there is a fade passing through the split point, recreate half of it on either side
		splitFadeLevel = self.GetFadeLevelAtPoint(self.selection[0])
		dictLeft[self.selection[0]] = splitFadeLevel
		dictRight[0.0] = splitFadeLevel
		
		millis = int(self.selection[0] * 1000)
		e.levels_list = self.levels_list.slice_by_endtime(millis)
		e.__fadePointsDict = dictRight
			
		e.__UpdateAudioFadePoints()
		e.SetProperties()
		self.instrument.events.append(e)
		e.emit("length")
		e.emit("position")
		
		return e
		

	
	@UndoSystem.UndoCommand("JoinEvent", "temp", "temp2")
	def SplitEvent(self, split_point, cutRightSide=True, eventID=-1):
		"""
		Splits this Event.
		
		Parameters:
			split_point -- time offset split_point in seconds to start the split.
			cutRightSide -- if True, a new event will be created to represent
					the piece on the right which was split. This instance
					will be the one on the left.
					if False, this instance is the one on the right.
			eventID -- the ID of an event that was previously used for the same
					split. This should only be used by the undo stack and will
					help prevent the creation of multiple garbage copies of the
					same event each time a split is undone and redone.
					
		Returns:
			the newly created event.
		"""
		dur = self.duration
		
		#if we were given an event ID, reuse that event instead of creating a new one
		if eventID >= 0:
			e = [x for x in self.instrument.graveyard if x.id == eventID][0]
			self.instrument.graveyard.remove(e)
		else:
			e = Event(self.instrument, self.file)
		e.name = self.name
		
		dictLeft = {}
		dictRight = {}
		for key, value in self.__fadePointsDict.iteritems():
			if key < split_point:
				dictLeft[key] = value
			if key > split_point:
				dictRight[key - split_point] = value
		#in case there is a fade passing through the split point, recreate half of it on either side
		splitFadeLevel = self.GetFadeLevelAtPoint(split_point)
		dictLeft[split_point] = splitFadeLevel
		dictRight[0.0] = splitFadeLevel
		
		if cutRightSide:
			e.start = self.start + split_point
			e.offset = self.offset + split_point
			e.duration = dur - split_point
			self.duration = split_point
			
			millis = int(split_point * 1000)
			e.levels_list = self.levels_list.slice_by_endtime(millis)
			self.levels_list = self.levels_list.slice_by_endtime(0, millis)
			
			self.__fadePointsDict = dictLeft
			e.__fadePointsDict = dictRight
		else:
			e.start = self.start
			e.offset = self.offset
			e.duration = split_point
		
			self.start = self.start + split_point
			self.offset = self.offset + split_point
			self.duration = dur - split_point
			
			millis = int(split_point * 1000)
			e.levels_list = self.levels_list.slice_by_endtime(0, millis)
			self.levels_list = self.levels_list.slice_by_endtime(millis)
			
			self.__fadePointsDict = dictRight
			e.__fadePointsDict = dictLeft
			
		self.__UpdateAudioFadePoints()
		e.__UpdateAudioFadePoints()
		e.SetProperties()
		self.instrument.events.append(e)
		self.SetProperties()
		self.emit("length")
		self.emit("position")
		self.instrument.emit("event::added", e)
		
		#undo parameters
		self.temp = e.id
		self.temp2 = cutRightSide
		
		return e
		
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("SplitEvent", "temp", "temp2", "temp3")
	def JoinEvent(self, joinEvent, joinToRight=True):
		"""
		Joins two Events together.
		
		Parameters:
			joinEvent -- the ID of the Event, or the Event object to join to this one.
			joinToRight -- True if the event will be joined on the right side
		"""
		eventObjectList = [x for x in self.instrument.events if x.id == joinEvent]
		if eventObjectList:
			joinEvent = eventObjectList[0]
		
		if joinToRight:
			self.temp = self.duration
			
			for key, value in joinEvent.__fadePointsDict.iteritems():
				self.__fadePointsDict[key + self.duration] = value
			#remove the point on either edge that was created when they were split
			if self.__fadePointsDict.has_key(self.duration):
				del self.__fadePointsDict[self.duration]
			
			old_duration = int(self.duration * 1000)
			self.duration += joinEvent.duration
			self.levels_list.extend(old_duration, joinEvent.levels_list)
			#update the fade point list after the level, and duration because it depends on them
			self.__UpdateAudioFadePoints()
		else:
			self.temp = joinEvent.duration
		
			self.start = joinEvent.start
			self.offset = joinEvent.offset
			
			old_duration = int(self.duration * 1000)
			self.duration += joinEvent.duration
			self.levels_list = LevelsList.add(old_duration, joinEvent.levels_list, self.levels_list)
			
			newDict = joinEvent.__fadePointsDict.copy()
			for key, value in self.__fadePointsDict.iteritems():
				newDict[key + joinEvent.duration] = value
			#remove the point on either edge that was created when they were split
			if newDict.has_key(joinEvent.duration):
				del newDict[joinEvent.duration]
			self.__fadePointsDict = newDict
			self.__UpdateAudioFadePoints()
			
		#create an undo action that is not attached to the project so that
		# the following delete will not be undone (it will be re-split not resurrected)
		nullAction = UndoSystem.AtomicUndoAction()
		# Now that they're joined, move delete the rightEvent
		joinEvent.Delete(_undoAction_=nullAction)
		
		self.emit("length")
		self.emit("position")
		
		self.temp2 = joinToRight
		self.temp3 = joinEvent.id

	#_____________________________________________________________________
	
	def Trim(self, start_split, end_split):
		"""
		Splits the Event and then deletes the first and last sections,
		leaving only the middle section.
		
		Parameters:
			start_split -- the time for the start of the trim.
			end_split -- the time for the end of the trim.
		"""
		
		if start_split > end_split or (start_split <= 0 and end_split >= self.duration):
			#both points must not be right at the edges, or there is nothing to split
			return
			
		undoAction = self.instrument.project.NewAtomicUndoAction()

		if 0 < start_split < self.duration:
			# Split off the left section of the event
			leftSplit = self.SplitEvent(start_split, False, _undoAction_=undoAction)
			self.instrument.DeleteEvent(leftSplit.id, _undoAction_=undoAction)
		
		#Adjust the end_split value since splitting the left has changed self.duration
		end_split = end_split - start_split
		
		if 0 < end_split < self.duration:
			# Split off the right section of the event
			rightSplit = self.SplitEvent(end_split, _undoAction_=undoAction)
			self.instrument.DeleteEvent(rightSplit.id, _undoAction_=undoAction)
		
		self.SetProperties()
		self.emit("length")
		
	#_____________________________________________________________________
	
	def _Compat02_UndoTrim(self, leftID, rightID):
		"""
		Resurrects two pieces from the graveyard and joins them to
		either side of this Event.
		
		Considerations: 
			This method is kept here for undo compatibility with Jokosher
			version 0.2 project files, and is never called when a 0.9 or
			greater project is being used.
		
		Parameters:
			leftID -- id of the left Event to be resurrected.
			rightID -- id of the right Event to be resurrected.
		"""
		undoAction = self.instrument.project.NewAtomicUndoAction()
		
		self.instrument.ResurrectEvent(leftID, _undoAction_=undoAction)
		self.JoinEvent(leftID, False, _undoAction_=undoAction)
		self.instrument.ResurrectEvent(rightID, _undoAction_=undoAction)
		self.JoinEvent(rightID, _undoAction_=undoAction)
		
	#_____________________________________________________________________
	
	def Delete(self, _undoAction_=None):
		"""
		Deletes this Event and sends it to the graveyard.
		"""
		if _undoAction_:
			self.instrument.DeleteEvent(self.id, _undoAction_=_undoAction_)
		else:
			self.instrument.DeleteEvent(self.id)
			
	#_____________________________________________________________________
	
	def Resurrect(self, _undoAction_=None):
		"""
		Brings this event back from the graveyard.
		
		Considerations:
			This method is made obsolete by instrument.ResurrectEvent(),
			but is still kept here for 0.2 undo history compatibility.
		"""
		if _undoAction_:
			self.instrument.ResurrectEvent(self.id, _undoAction_=_undoAction_)
		else:
			self.instrument.ResurrectEvent(self.id)
			
	#_____________________________________________________________________

	def install_plugin_cb(self, result):
		self._installing_plugins = False
		# hardcode gst.pbutils.INSTALL_PLUGINS_SUCCESS to avoid conditional gst.pbutils import
		INSTALL_PLUGINS_SUCCESS = 0
		if result == INSTALL_PLUGINS_SUCCESS:
			gst.update_registry()
			self.GenerateWaveform()
			return

		# FIXME: send a better error
		msg = "failed to install plugins: %s" % result
		self.emit("corrupt", msg)

	#_____________________________________________________________________
	
	def bus_message(self, bus, message):
		"""
		Handler for the GStreamer bus messages relevant to this Event.
		At the moment, this is used to report on how the loading progress
		is going.
		
		Parameters:
			bus -- GStreamer bus sending the message.
			message -- GStreamer message.
			
		Returns:
			True -- the Event is loading.
			False -- the Event isn't loading.
		"""
		if not self.isLoading:
			return False

		st = message.structure
		if not st:
			return False
		
		if st.get_name().startswith('missing-'):
			self.loadingPipeline.set_state(gst.STATE_NULL)
			Utils.HandleGstPbutilsMissingMessage(message, self.install_plugin_cb)

		elif st.get_name() == "level":
			self.__AppendLevelToList(st)
			
			#Truncate so it updates once per second
			self.loadingLength = st["endtime"] / gst.SECOND
			
			# Only send events every second processed to reduce GUI load
			if self.loadingLength != self.lastEnd:
				self.lastEnd = self.loadingLength 
				self.emit("length") # tell the GUI
		return True
		
	#_____________________________________________________________________
		
	def bus_eos(self, bus, message):	
		"""
		Handler for the GStreamer End Of Stream message. Currently
		used when the file is loading and is being rendered. This
		function is called at the end of the file loading process and
		finalises the rendering.
		
		Parameters:
			bus -- GStreamer bus sending the message.
			message -- GStreamer message.
			
		Returns:
			False -- stops the signal propagation. *CHECK*
		"""
		if message.type == gst.MESSAGE_EOS:
			
			# Update levels for partial events
			q = self.loadingPipeline.query_duration(gst.FORMAT_TIME)
			length = float(q[0] / float(gst.SECOND))
			
			#we're at EOS, and still have no value for duration
			if not self.duration:
				if length:
					self.duration = length
				else:
					self.duration = self.loadingLength
			
			if self.levels_list:
				final_endtime = self.levels_list[-1][0]
				if final_endtime > int(self.duration * 1000):
					Globals.debug("Event %d: duration (%f) is less than last level endtime (%d)."
					              % (self.id, self.duration, final_endtime))
					self.duration = final_endtime / 1000.0
					self.SetProperties()
					Globals.debug("\tduration has been increased to", self.duration)
			
			if length and (self.offset > 0 or self.duration != length):
				starttime = int(self.offset * 1000)
				stoptime = int((self.offset + self.duration) * 1000)
				self.levels_list = self.levels_list.slice_by_endtime(starttime, stoptime)
				
			# We're done with the bin so release it
			self.StopGenerateWaveform()
			
			# Signal to interested objects that we've changed
			self.emit("waveform")
			return False
			
	#_____________________________________________________________________

	def bus_message_statechange(self, bus, message):
		"""
		Handler for the GStreamer statechange message.
		
		Parameters:
			bus -- GStreamer bus sending the message.
			message -- GStreamer message.
		"""
		# state has changed
		try:
			time = self.loadingPipeline.query_duration(gst.FORMAT_TIME)
			if self.duration == 0 and time[0] > 0:
				self.duration = float(time[0] / float(gst.SECOND))
				
				#update position with proper duration
				self.MoveButDoNotOverlap(self.start)
				self.SetProperties()
				self.emit("length")
				self.emit("position")
		except:
			# no size available yet
			pass

	#_____________________________________________________________________

	def bus_error(self, bus, message):
		"""
		Handler for when things go completely wrong with GStreamer.
		
		Parameters:
			bus -- GStreamer bus sending the message.
			message -- GStreamer message.
		"""
		error, debug = message.parse_error()
		
		Globals.debug("Event Bus Error Message:")
		Globals.debug("\tCode:", error.code)
		Globals.debug("\tDomain:", error.domain)
		Globals.debug("\tMessage:", error.message)
		
		Globals.debug("Event bus error:", str(error), str(debug))
		self.emit("corrupt", "%s\n%s" % (error, debug))
	
	#_____________________________________________________________________
	
	def bus_message_tags(self, bus, message):
		"""
		Handler for catching audio file tags that Gstreamer throws.
		
		Parameters:
			bus -- GStreamer bus sending the message.
			message -- GStreamer message.
		"""
		Globals.debug("received group of tags")
		st = message.structure
		
		title, artist = None, None
		if st.has_key("title"):
			title = st["title"]
		if st.has_key("artist"):
			artist = st["artist"]
		
		#check which tags actually contain something
		if title and artist:
			self.name = _("%(title)s by %(artist)s") % {"title":title, "artist":artist}
		elif title:
			self.name = title
	
	#_____________________________________________________________________
	
	def GenerateWaveform(self):
		"""
		Renders the level information for the GUI.
		"""
		pipe = """filesrc name=src ! decodebin ! audioconvert ! level message=true name=level_element ! fakesink"""
		self.loadingPipeline = gst.parse_launch(pipe)
		
		filesrc = self.loadingPipeline.get_by_name("src")
		level = self.loadingPipeline.get_by_name("level_element")
		
		filesrc.set_property("location", self.GetAbsFile())
		level.set_property("interval", int(self.LEVEL_INTERVAL * gst.SECOND))

		self.bus = self.loadingPipeline.get_bus()
		self.bus.add_signal_watch()
		self.bus.connect("message::element", self.bus_message)
		self.bus.connect("message::tag", self.bus_message_tags)
		self.bus.connect("message::state-changed", self.bus_message_statechange)
		self.bus.connect("message::eos", self.bus_eos)
		self.bus.connect("message::error", self.bus_error)

		self.levels_list = LevelsList.LevelsList()
		self.isLoading = True
		self.emit("loading")

		self.loadingPipeline.set_state(gst.STATE_PLAYING)

	#_____________________________________________________________________
	
	def CopyAndGenerateWaveform(self, uri):
		"""
		Copies the audio file to the new file location and reads the levels
		at the same time.
		"""
		
		urisrc = gst.element_make_from_uri(gst.URI_SRC, uri)
		if not urisrc:
			#This means that here is no gstreamer src element on the system that can handle this URI type.
			return False
		
		pipe = """tee name=mytee mytee. ! queue ! filesink name=sink """ +\
		       """mytee. ! queue ! decodebin ! audioconvert ! level name=level_element message=true ! fakesink""" 
		self.loadingPipeline = gst.parse_launch(pipe)
		
		tee = self.loadingPipeline.get_by_name("mytee")
		filesink = self.loadingPipeline.get_by_name("sink")
		level = self.loadingPipeline.get_by_name("level_element")
		
		self.loadingPipeline.add(urisrc)
		urisrc.link(tee)
		
		filesink.set_property("location", self.GetAbsFile())
		level.set_property("interval", int(self.LEVEL_INTERVAL * gst.SECOND))

		self.bus = self.loadingPipeline.get_bus()
		self.bus.add_signal_watch()
		self.bus.connect("message::element", self.bus_message)
		self.bus.connect("message::tag", self.bus_message_tags)
		self.bus.connect("message::state-changed", self.bus_message_statechange)
		self.bus.connect("message::eos", self.bus_eos)
		self.bus.connect("message::error", self.bus_error)

		self.levels_list = LevelsList.LevelsList()
		self.isLoading = True
		self.emit("loading")

		self.loadingPipeline.set_state(gst.STATE_PLAYING)
		
		return True
		
	#_____________________________________________________________________
	
	def StopGenerateWaveform(self, finishedLoading=True):
		"""
		Stops the internal pipeline that loads the waveform from this event's file.
		
		Parameters:
			finishedLoading -- True if the event has finished loading the waveform,
					False if the loading is being cancelled.
		"""
		if self.bus:
			self.bus.remove_signal_watch()
			self.bus = None
		if self.loadingPipeline:
			self.loadingPipeline.set_state(gst.STATE_NULL)
			
			if finishedLoading and self.levels_list:
				self.levels_list.tofile(self.GetAbsLevelsFile())
				del_on_close_list = self.instrument.project.deleteOnCloseAudioFiles
				# this event might not be in the project file yet
				# if so, levels_file should be deleted when audio file is deleted on exit
				if self.GetAbsFile() in del_on_close_list:
					del_on_close_list.append(self.GetAbsLevelsFile())
					
				inc = IncrementalSave.CompleteLoading(self.id, self.duration, self.levels_file)
				self.instrument.project.SaveIncrementalAction(inc)
			
			if self.isDownloading:
				# If we are currently downloading, we can't restart later, 
				# so cancel regardless of the finishedLoading boolean's value.
				self.isDownloading = False
				self.isLoading = False
			else:
				self.isLoading = not finishedLoading
			
			
			self.loadingPipeline = None
			self.loadingLength = 0
			self.emit("loading")
	
	#_____________________________________________________________________

	def recording_bus_level(self, bus, message):
		"""
		Handler for the GStreamer bus messages relevant to this Event.
		At the moment this is used to report on how the recording
		progress is going. *CHECK*
		
		Parameters:
			bus -- GStreamer bus sending the message.
			message -- GStreamer message.
			
		Returns:
			True -- the Event is recording.
			False -- the Event isn't recording.
		"""
		if not self.isRecording:
			return False
		
		st = message.structure
		if st and message.src.get_name() == "recordlevel":
			self.__AppendLevelToList(st)
			
			end = st["endtime"] / float(gst.SECOND)  #convert to float representing seconds 
			#Round to one decimal place so it updates 10 times per second
			self.loadingLength = round(end, 1)
			
			# Only send events every second processed to reduce GUI load
			if self.loadingLength != self.lastEnd:
				self.lastEnd = self.loadingLength 
				self.emit("length") # tell the GUI
		return True
		
	#_____________________________________________________________________
	
	def __AppendLevelToList(self, structure):
		(end, peaks) = Utils.CalculateAudioLevelFromStructure(structure)
		
		# the last level may be sent twice. If timestamp is the same, ignore it.
		if self.levels_list and self.levels_list[-1][0] == end:
			return
		
		# work around GStreamer bug where stream time will be -1 (indicating error)
		# and then cast to guint64 which results in the maximum 64-bit integer value.
		# In this case stream-time and endtime are bogus values, but duration is still correct.
		stream_time = structure["stream-time"]
		if stream_time == ((2**64) - 1):
			delta = int(structure["duration"] / Utils.NANO_TO_MILLI_DIVISOR)
			self.levels_list.append_time_delta(delta, peaks)
		else:
			self.levels_list.append(end, peaks)
		
	#_____________________________________________________________________
	
	def SetSelected(self, sel):
		"""
		Enables or disables the selection state for this Event.

		Parameters:
			sel -- the new selection state:
					True = the Event has been selected.
					False = the Event has been deselected.
		"""
		# No need to emit a signal when there is no change in selection state
		if self.isSelected is not sel:
			self.isSelected = sel
			self.emit("selected")
	
	#_____________________________________________________________________
	
	def MayPlace(self, xpos):
		"""
		Checks if this event could be placed at xpos without 
		overlapping another Event on the same Instrument.

		Parameters:
			xpos -- the potential start position to check.
		
		Returns:
			True if it's OK to place the Event at xpos, False if not.
		"""
		for e in self.instrument.events:
			if e is self:
				continue
			if not (e.start + e.duration <= xpos or e.start >= xpos + self.duration):
				return False
		return True
		
	#_____________________________________________________________________
	
	def MoveButDoNotOverlap(self, xpos):
		"""
		This method will attempt to move this Event to the given position.
		If the position requires overlapping, this Event will be put flush
		against the closest side of the Event which is in the way.

		Parameters:
			xpos -- the potential position to move the Event to.
		"""
		alreadyTriedRemovingOverlap = False
		self.instrument.events.sort()
		
		for e in self.instrument.events:
			if e is self:
				continue
			
			elif alreadyTriedRemovingOverlap:
				#we have already attempted to place to the left, on the last iteration
				#from now on we only try placing on the right side
				start = e.start + e.duration
				if self.MayPlace(start):
					self.start = start
					return
				continue
			
			rightPos = e.start + e.duration
			leftPos = e.start - self.duration
			#if this event IS overlapping with the other event
			if not (rightPos <= xpos or leftPos >= xpos):
				#if the middle of this event is on the RIGHT of the middle of the other event
				if rightPos > xpos and e.start + (e.duration/2) < xpos + (self.duration/2):
					order = (rightPos, max(leftPos, 0))
				else: #if the middle is on the LEFT
					order = (max(leftPos, 0), rightPos)
				
				for start in order:
					if self.MayPlace(start):
						self.start = start
						return
				
				alreadyTriedRemovingOverlap = True
				
		#There are no other events overlapping with this one
		self.start = xpos
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("RemoveAudioFadePoints", "temp", "temp2", "temp3", "temp4")
	def AddAudioFadePoints(self, firstPoint, secondPoint, firstVolume, secondVolume):
		"""
		Adds two fade points to the audioFadePoints list.
		If either point exists already, replaces it, and resorts the list by time.
		
		Parameters:
			firstPoint -- point in time, where the fade starts.
			secondPoint -- point in time, where the fade ends.
			firstVolume -- value of the initial fade volume.
			secondVolume -- value of the final fade volume.
		"""
		#for command manager to use with undo
		self.temp = firstPoint
		self.temp2 = secondPoint
		self.temp3 = None
		self.temp4 = None
		
		if self.__fadePointsDict.has_key(firstPoint):
			self.temp3 = self.__fadePointsDict[firstPoint]
		if self.__fadePointsDict.has_key(secondPoint):
			self.temp4 = self.__fadePointsDict[secondPoint]
		
		#we *must* compare to None here because audio points with volume 0 *are* allowed
		if firstPoint != None and firstVolume != None:
			self.__fadePointsDict[firstPoint] = firstVolume
		if secondPoint != None and secondVolume != None:
			self.__fadePointsDict[secondPoint] = secondVolume
		
		self.__UpdateAudioFadePoints()
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("AddAudioFadePoints", "temp", "temp2", "temp3", "temp4")
	def RemoveAudioFadePoints(self, firstPoint, secondPoint, firstOldVolume=None, secondOldVolume=None):
		"""
		Removes a fade point (along with its values) from the audioFadePoints list.
		If firstOldVolume and secondOldVolume are given, the levels at the two points
		will be replaced with those old values, instead of being removed.
		
		Parameters:
			firstPoint -- point in time, where the fade starts.
			secondPoint -- point in time, where the fade ends.
			firstOldVolume -- old value of the initial fade volume,
								used to replace the current one.
			secondOldVolume -- old value of the final fade volume,
								used to replace the current one.
		"""
		#undo values
		self.temp = firstPoint
		self.temp2 = secondPoint
		self.temp3 = None
		self.temp4 = None
		
		if self.__fadePointsDict.has_key(firstPoint):
			self.temp3 = self.__fadePointsDict[firstPoint]
		if self.__fadePointsDict.has_key(secondPoint):
			self.temp4 = self.__fadePointsDict[secondPoint]
		
		#if the point had a previous value (firstOldPoint) then put the value back
		#we *must* compare to None here because audio points with volume 0 *are* allowed
		if firstOldVolume != None:
			self.__fadePointsDict[firstPoint] = firstOldVolume
		#else, just remove the point because it didn't exist before
		elif self.__fadePointsDict.has_key(firstPoint):
			del self.__fadePointsDict[firstPoint]
		
		#same as above but for the second point
		if secondOldVolume != None:
			self.__fadePointsDict[secondPoint] = secondOldVolume
		elif self.__fadePointsDict.has_key(secondPoint):
			del self.__fadePointsDict[secondPoint]
			
		self.__UpdateAudioFadePoints()
	
	#_____________________________________________________________________
	
	def __UpdateAudioFadePoints(self):
		"""
		Private function that uses the private dictionary with
		all the fade points to update the audioFadePoints list.
		The audioFadePoints is different from the dictionary
		because it will always have points at the beginning and
		the end of the event (unless there are none at all), and
		it is a list, so accessing the contents in order is much faster.
		"""
		
		#update the fade points list from the dictionary
		self.audioFadePoints = self.__fadePointsDict.items()
		#dicts dont have order, so sort after update
		self.audioFadePoints.sort(key=lambda x: x[0])
		
		#only add beginning and end points if there are some other points already in the list
		if self.audioFadePoints:
			#if there is no point at the beginning, make one with the
			#same value as the first fade point. This makes the first fade
			#extend back to the beginning.
			if self.audioFadePoints[0][0] != 0.0:
				first = (0.0, self.audioFadePoints[0][1])
				self.audioFadePoints.insert(0, first)
			#same as above but for the end of the event.
			if self.audioFadePoints[-1][0] != self.duration:
				last = (self.duration, self.audioFadePoints[-1][1])
				self.audioFadePoints.append(last)
			
		self.__UpdateFadeLevels()
		self.emit("waveform")
	
	#_____________________________________________________________________
	
	def __UpdateFadeLevels(self):
		"""
		Private function that uses the private dictionary with
		all the fade leves to update the fadeLevels list. The fadeLevels
		list is a cache of faded levels as they will be shown on the screen
		so that we don't have to calculate them everytime we draw.
		"""
		if not self.audioFadePoints or len(self.audioFadePoints) < 2:
			Globals.debug("Event", self.id, ": no fade points to use")
			#there are no fade points for us to use
			return
			
		#fadePercents = []
		#oneSecondInLevels = len(self.levels) / self.duration
		
		self.fadeLevels = LevelsList.LevelsList()
		
		iterFadePoints = iter(self.audioFadePoints)
		firstFadeTime, firstFadeValue = iterFadePoints.next()
		firstFadeTime = int(firstFadeTime * 1000)	#convert to milliseconds
		secondFadeTime, secondFadeValue = iterFadePoints.next()
		secondFadeTime = int(secondFadeTime * 1000)	#convert to milliseconds
		# if less than one percent difference, assume they are the same
		sameValues = abs(firstFadeValue - secondFadeValue) < 0.01
		
		if not sameValues:
			slope = (secondFadeValue - firstFadeValue) / (secondFadeTime - firstFadeTime)
		
		for endtime, peak in self.levels_list:
			# check if we have moved into the next fade point pair
			# don't care about 1 millisecond difference, its rounding error
			if endtime > (secondFadeTime + 1):  
				firstFadeTime = secondFadeTime
				firstFadeValue = secondFadeValue
				try:
					secondFadeTime, secondFadeValue = iterFadePoints.next()
				except StopIteration:
					Globals.debug("Event %d: endtime (%d) is after last fade point (%d,%d)"
					              % (self.id, endtime, secondFadeTime, secondFadeValue))
				
				secondFadeTime = int(secondFadeTime * 1000)	#convert to milliseconds
				
				# if less than one percent difference, assume they are the same
				sameValues = abs(firstFadeValue - secondFadeValue) < 0.01
				if not sameValues:
					# the fade line is not flat, so calculate the slope of it
					slope = (secondFadeValue - firstFadeValue) / (secondFadeTime - firstFadeTime)
			
			if sameValues:
				#no fade here, the same volume continues across
				self.fadeLevels.append(endtime, [int(peak * firstFadeValue)])
			else:
				rel_time = endtime - firstFadeTime
				peak_delta = slope * rel_time
				new_fade_value = firstFadeValue + peak_delta
				assert new_fade_value <= 1.0		# a fade cannot be more than 100%
				peak = int(peak * new_fade_value)
				
				self.fadeLevels.append(endtime, [peak])
	#_____________________________________________________________________
	
	def GetFadeLevels(self):
		"""
		Obtain the fade levels list.
		The only difference between this list and the levels list is
		that the levels in this list are scaled according to the fade
		curves applied to the current Event.
		
		Returns:
			a list of fade levels, the same length as the levels list.
		"""
		# no fades registered
		if not self.audioFadePoints:
			return self.levels_list
			
		if len(self.fadeLevels) != len(self.levels_list):
			self.__UpdateFadeLevels()
			assert len(self.fadeLevels) == len(self.levels_list)
			
		return self.fadeLevels
		
	#_____________________________________________________________________
	
	def GetFadeLevelAtPoint(self, time):
		"""
		Obtain the level of audio at any point in time.
		
		Parameters:
			time -- point in time to extract the audio level from.
		
		Returns:
			the level of the audio in percentage format [0,1]
		"""
		if not self.audioFadePoints:
			return 1.0
		if self.audioFadePoints[-1][0] < time or self.audioFadePoints[0][0] > time:
			#for some reason the time given is outside the event, so ignore it
			return 1.0
		
		#we can assume that audioFadePoints is sorted and has at least 2 elements
		points = self.audioFadePoints
		for i in xrange(1, len(points)):
			if points[i][0] >= time:
				right = points[i]
				left = points[i-1]
				break
		
		if right[0] == left[0]:
			return 1.0
		elif right[1] == left[1]:
			return left[1]
		else:
			ratio = (right[1] - left[1]) / (right[0] - left[0])
			
		relativeLevel = ratio * (time - left[0])
		return left[1] + relativeLevel
		
	#_____________________________________________________________________
	
	def DeleteSelectedFadePoints(self):
		"""
		Removes all the fade points for this Event.
		"""
		removeList = []
		for key in self.__fadePointsDict.iterkeys():
			if self.selection[0] <= key <= self.selection[1]:
				removeList.append(key)
		
		if len(removeList) % 2 == 0:
			#if we have an even number, use the fact that
			#RemoveAudioFadePoints takes two points.
			for i in range(len(removeList))[::2]:
				self.RemoveAudioFadePoints(removeList[i], removeList[i+1])
		else:
			for i in removeList:
				self.RemoveAudioFadePoints(i, None)
	
	#_____________________________________________________________________
#=========================================================================	



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	WelcomeDialog.py
#	
#	This class displays the welcome dialog when the application starts.
#
#-------------------------------------------------------------------------------

import gtk.glade
import os
import Globals
import PreferencesDialog

#=========================================================================

class WelcomeDialog:
	"""
	This class handles all of the processing associated with the
	Welcome Dialog (which comes up on start).
	"""	
	#_____________________________________________________________________

	def __init__(self, mainwindow):
		"""
		Creates a new instance of MainApp.
		
		Parameters:
			mainwindow -- instance of JokosherApp. Used for parent/child relationship.
		"""

		# this points to JokosherApp		
		self.mainwindow = mainwindow
				
		self.res = gtk.glade.XML(Globals.GLADE_PATH, "WelcomeDialog")

		self.signals = {
			"on_newproject_clicked" : self.OnNewProject,
			"on_openproject_clicked" : self.OnOpenProject,
			"on_preferences_clicked" : self.OnPreferences,
			"on_openrecentprojectbutton_clicked" : self.OnOpenRecentProjectButton,
			"on_quit_clicked" : self.OnQuit,
			"on_startupcb_toggled" : self.OnStartupToggled,
		}
		
		self.res.signal_autoconnect(self.signals)

		self.window = self.res.get_widget("WelcomeDialog")
		self.window.set_icon(self.mainwindow.icon)
		self.window.set_transient_for(self.mainwindow.window)
		
		self.tree = self.res.get_widget("recentprojectslist")
		self.headerimage = self.res.get_widget("headerimage")
		self.headerimage.set_from_file(os.path.join(Globals.IMAGE_PATH, "welcomeheader.png"))

		self.openrecentbutton = self.res.get_widget("openrecentprojectbutton")
		self.openrecentbutton.set_sensitive(False)

		# set up recent projects treeview with a ListStore model. We also
		# use CellRenderPixbuf as we are using icons for each entry
		self.model = gtk.ListStore(str, str, str)
		self.PopulateRecentProjects()
		self.tree.set_model(self.model)
		self.tvcolumn = gtk.TreeViewColumn('Recent Projects')
		self.cellpb = gtk.CellRendererPixbuf()
		self.cell = gtk.CellRendererText()
		
		self.tvcolumn.pack_start(self.cellpb, False)
		self.tvcolumn.pack_start(self.cell, True)
		
		self.tvcolumn.set_attributes(self.cellpb, stock_id=0)
		self.tvcolumn.set_attributes(self.cell, text=1)
		
		self.tree.append_column(self.tvcolumn)

		self.tree.connect("row-activated", self.OnRecentProjectSelected)
		self.tree.connect("cursor-changed", self.OnEnableRecentProjectButton)
		
		self.window.show_all()

	#_____________________________________________________________________
	
	def OnNewProject(self, widget):
		"""
		Starts a new project.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		# hide the welcome dialog and call OnNewProject() from JokosherApp
		self.window.hide()
		self.mainwindow.OnNewProject(self, self.OnDialogClose)
	
	#_____________________________________________________________________
			
	def OnOpenProject(self, button=None):
		"""
		Opens an existing project.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""

		# hide the welcome dislog and call OnOpenProject() from JokosherApp		
		self.window.hide()
		self.mainwindow.OnOpenProject(self, self.OnDialogClose)
		
	#_____________________________________________________________________
	
	def OnPreferences(self, button):
		"""
		Shows the preferences window.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""

		# hide the welcome dislog and call OnPreferences() from JokosherApp		
		self.window.hide()
		self.mainwindow.OnPreferences(self, self.OnDialogClose)
		
	#_____________________________________________________________________

	def OnDialogClose(self, dialog=None):
		"""
		The dialog was closed.
		
		Parameters:
			dialog -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		if self.mainwindow.project:
			self.window.destroy()
		else:
			self.window.show_all()
		
	#_____________________________________________________________________

	def OnQuit(self, button):
		"""
		Quits Jokosher.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		gtk.main_quit()

	#_____________________________________________________________________

	def PopulateRecentProjects(self):
		"""
		Populate the Recent Projects menu with items from global settings.
		"""
		self.model.clear()
		for path, name in self.mainwindow.recentprojectitems:	
			self.model.append([gtk.STOCK_NEW, name, path])
	
	#_____________________________________________________________________

	def OnRecentProjectSelected(self, treeview, path, view_column):
		"""
		This method is called when one of the entries in the recent projects
		list is selected.
		
		Parameters:
			treeview -- reserved for GTK callbacks, don't use it explicitly.
			path -- path to the project file.
			view_column -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		item = self.model[path]
		response = self.mainwindow.OnRecentProjectsItem(self, item[2], item[1])
		if response:
			#it opened without error, so close our window
			self.window.destroy()
		
	#_____________________________________________________________________

	def OnEnableRecentProjectButton(self, treeview):
		"""
		When a recent project is selected, enable the button to load it.
		
		Parameters:
			treeview -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		self.openrecentbutton.set_sensitive(True)
		
	#_____________________________________________________________________

	def OnOpenRecentProjectButton(self, widget):
		"""
		Loads the selected recent project.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		item = self.model[self.tree.get_cursor()[0]]
		self.mainwindow.OnRecentProjectsItem(self, item[2], item[1])
		self.window.destroy()
	
	#_____________________________________________________________________

	def OnStartupToggled(self, widget):
		"""
		When the startup toggle box is toggled, this method is run to update
		the setting in Globals.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		if widget.get_active():
			Globals.settings.general["startupaction"] = PreferencesDialog.STARTUP_NOTHING
		else:	
			Globals.settings.general["startupaction"] = PreferencesDialog.STARTUP_WELCOME_DIALOG
		
		Globals.settings.write()
	
	#_____________________________________________________________________
#=========================================================================




import urllib


def getFullName():
	#TODO: Work out how to get the fullname in windows
	return ""

def samefile(path1, path2):
	return path1 == path2

def url2pathname(url):
	return urllib.url2pathname(url)

def pathname2url(path):
	#Windows pathname2url appends // to the front of the path
	return "file:%s" % urllib.pathname2url(path)

def GetRecordingDefaults():
	defaults = {
			"fileformat": "vorbisenc ! oggmux",
			"file_extension": "ogg",
			"audiosrc" : "dshowaudiosrc"
		}
	return defaults


def GetPlaybackDefaults():
	defaults = {
			"audiosink": "directsoundsink"
		}
	return defaults	



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Utils.py
#	
#	This module contains a bunch of useful helper methods used elsewhere in
#	the code, that are not specific to any particular class.
#
#-------------------------------------------------------------------------------

import xml.dom.minidom as xml
import math, os.path, sys
import gtk, gobject
import webbrowser
import Globals

import gst
try:	
	import gst.pbutils
	have_pbutils = True
except:
	have_pbutils = False

# the highest range in decibels there can be between any two levels
DECIBEL_RANGE = 80

NANO_TO_MILLI_DIVISOR = gst.SECOND / 1000

#_____________________________________________________________________

def OpenExternalURL(url, message, parent, timestamp=0):
	"""
	Opens a given url in the user's default web browser.
		
	Parameters:
		url -- the url the user's default web browser will open.
		message -- the error message in the dialog window. the error message dialog will show if a link cannot be opened.
		parent -- parent window of the error message dialog.
	"""
		
	screen = gtk.gdk.screen_get_default()
	ret = False
	try:
		ret = gtk.show_uri(screen, url, timestamp)
	except:
		ret = webbrowser.open(url)
	
	if not ret and message:
		dlg = gtk.MessageDialog(parent,
				gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
				gtk.MESSAGE_ERROR,
				gtk.BUTTONS_CLOSE)
		dlg.set_markup(message % url)
		dlg.run()
		dlg.destroy()
	
	return ret
#_____________________________________________________________________

def GetIconThatMayBeMissing(iconName, iconSize, returnGtkImage=True):
	"""
	First tries to get the icon with the given name from the icon theme.
	If that fails, it will try and load it from the Jokosher image directory.
	
	Parameters:
		iconName -- the string of the icon's name
		iconSize -- an icon size from the gtk.
		returnGtkImage -- If True, this method will return a gtk.Image.
				If False it will return a gtk.gdk.Pixbuf.
	
	Returns:
		an instance of gtk.gdk.Pixbuf that contains the requested image.
	"""
	pixbuf = None
	try:
		pixbuf = gtk.icon_theme_get_default().load_icon(iconName, iconSize, 0)
	except gobject.GError:
		path = os.path.join(Globals.IMAGE_PATH, "%s.png" % iconName)
		try:
			pixbuf = gtk.gdk.pixbuf_new_from_file(path)
		except:
			pass
	
	if not returnGtkImage:
		return pixbuf
	
	if pixbuf:
		image = gtk.Image()
		image.set_from_pixbuf(pixbuf)
	else:
		image = gtk.image_new_from_stock(gtk.STOCK_MISSING_IMAGE, iconSize)
	
	return image

#_____________________________________________________________________

def DbToFloat(f):
	"""
	Converts f from the decibel scale to a float.
	
	Parameters:
		f -- number in decibel format.
	
	Returns:
		a float in the [0,1] range.
	"""
	return pow(10., f / 20.)

#_____________________________________________________________________

def CalculateAudioLevelFromStructure(structure):
	"""
	Calculates an average for all channel levels.
	
	Parameters:
		channelLevels -- list of levels from each channel.
		
	Returns:
		an average level, also taking into account negative infinity numbers,
		which will be discarded in the average.
	"""
	# FIXME: currently everything is being averaged to a single channel
	channelLevels = structure["rms"]

	negInf = -1E+5000
	peaktotal = 0
	for peak in channelLevels:
		#if peak > 0.001:
		#	print channelLevels
		#don't add -inf values cause 500 + -inf is still -inf
		if peak != negInf:
			peaktotal += peak
		else:
			peaktotal -= DECIBEL_RANGE
	
	peaktotal /= len(channelLevels)
	
	peaktotal += DECIBEL_RANGE
	#convert to an integer
	peaktotal = min(peaktotal, DECIBEL_RANGE)
	peaktotal = max(peaktotal, -DECIBEL_RANGE)
	peakint = int((peaktotal / DECIBEL_RANGE) * sys.maxint)

	endtime = structure["endtime"]
	#convert number from gst.SECOND (i.e. nanoseconds) to milliseconds
	endtime_millis = int(endtime / NANO_TO_MILLI_DIVISOR)

	return (endtime_millis, [peakint])

#_____________________________________________________________________

def floatRange(start, end=None, inc=None):
	"""
	A range function capable of performing float increments.
	
	Parameters:
		start -- minimum value of the range list.
		end -- maximum value of the range list.
		inc -- delta between values in the range list.
	
	Considerations:
		If start is the only parameter given, the range goes from [0.0, start].
	
	Returns:
		A list with the range [start, end] in inc increments.
	"""
	if end == None:
		end = float(start)
		start = 0.0
	else:
		start = float(start)
	if not inc:
		inc = 1.0
	#check if the increment has the wrong sign
	#if it does, it may decrement instead of increment and we will get an infinite loop
	elif (start > end and inc > 0) or (start <= end and inc < 0):
		inc = -inc
		
	count = int(math.ceil((end - start) / inc))
	L = [None,] * max(count, 1)

	L[0] = start
	for i in xrange(1, count):
		L[i] = L[i-1] + inc
		
	return L

#_____________________________________________________________________

def StoreParametersToXML(self, doc, parent, parameters):
	"""
	Saves the variables indicated by the parameters in an XML document.
	
	Parameters:
		doc -- name of the XML document to save the settings into.
		parent -- XML parent tag to use in doc.
		parameters -- list of variable names whose value, save in doc.
	"""	   
	for param in parameters:
		node = doc.createElement(param)
		StoreVariableToNode(getattr(self, param), node)
		parent.appendChild(node)

#_____________________________________________________________________

def LoadParametersFromXML(self, parentElement):
	"""
	Loads parameters from an XML and fills variables of the same name
	in that module.
	
	Parameters:
		parentElement -- block of XML with the parameters.
	"""
	for node in parentElement.childNodes:
		if node.nodeType == xml.Node.ELEMENT_NODE:
			value = LoadVariableFromNode(node)
			setattr(self, node.tagName, value)
			
#_____________________________________________________________________

def StoreDictionaryToXML(doc, parent, dict, tagName=None):
	"""
	Saves a dictionary of settings in an XML document.
	
	Parameters:
		doc -- name of the XML document to save the settings into.
		parent -- XML parent tag to use in doc.
		dict -- dictionary to be saved in doc.
		tagName -- name used for all tag names.
		
	Considerations:
		If tagName is not given, the dictionary keys will be used for the tag names.
		This means that the keys must all be strings and can't have any invalid XML
		characters in them.
		If tagName is given, it is used for all the tag names, and the key is stored
		in the keyvalue attribute and its type in the keytype attribute.
	"""
	for key, value in dict.iteritems():
		if tagName:
			node = doc.createElement(tagName)
			StoreVariableToNode(key, node, "keytype", "keyvalue")
		#if no tag name was provided, use the key
		else:
			node = doc.createElement(key)
		
		StoreVariableToNode(value, node, "type", "value")
		parent.appendChild(node)

#_____________________________________________________________________

def LoadDictionaryFromXML(parentElement):
	"""
	For those times when you don't want to fill module variables with
	parameters from the XML but just want to fill a dictionary instead.
	
	Parameters:
		parentElement -- XML element from which the dictionary is loaded.
	
	Returns:
		a dictionary with the loaded values in (type, value) format.
	"""
	dictionary = {}
	
	for node in parentElement.childNodes:
		if node.nodeType == xml.Node.ELEMENT_NODE:
			if node.hasAttribute("keytype") and node.hasAttribute("keyvalue"):
				key = LoadVariableFromNode(node, "keytype", "keyvalue")
			else:
				key = node.tagName
			value = LoadVariableFromNode(node, "type", "value")
			dictionary[key] = value
	
	return dictionary

#_____________________________________________________________________

def StoreListToXML(doc, parent, itemList, tagName):
	"""
	Saves a list of items in an XML document.
	
	Parameters:
		doc -- name of the XML document to save the items into.
		parent -- XML parent tag to use in doc.
		itemList -- list of items to be saved in doc.
		tagName -- name used for all tag names.
	"""
	for value in itemList:
		node = doc.createElement(tagName)
		StoreVariableToNode(value, node)
		parent.appendChild(node)

#_____________________________________________________________________

def LoadListFromXML(parentElement):
	"""
	Loads a list from an XML file.
	
	Parameters:
		parentElement -- block of XML with the list nodes.
		
	Returns:
		a list with the loaded values.
	"""
	itemList = []
	
	for node in parentElement.childNodes:
		if node.nodeType == xml.Node.ELEMENT_NODE:
			value = LoadVariableFromNode(node)
			itemList.append(value)
	
	return itemList

#_____________________________________________________________________

def LoadVariableFromNode(node, typeAttr="type", valueAttr="value"):
	"""
	Loads a variable from an specific XML node.
	
	Example:
		Please refer to the StoreVariableToNode example
		for the explanation of the typeAttr and valueAttr
		parameters.
	
	Parameters:
		node -- node from which the variable is loaded.
		typeAttr -- string of the attribute name that the
					variable's type will be saved under.
		valueAttr -- string of the attribute name that the
					variable's value will be saved under.
	
	Returns:
		the loaded variable.
	"""
	if node.getAttribute(typeAttr) == "int":
		variable = int(node.getAttribute(valueAttr))
	elif node.getAttribute(typeAttr) == "float":
		variable = float(node.getAttribute(valueAttr))
	elif node.getAttribute(typeAttr) == "bool":
		variable = (node.getAttribute(valueAttr) == "True")
	elif node.getAttribute(typeAttr) == "NoneType":
		variable = None
	else:
		variable = node.getAttribute(valueAttr)
	
	return variable

#_____________________________________________________________________

def StoreVariableToNode(value, node, typeAttr="type", valueAttr="value"):
	"""
	Saves a variable to an specific XML node.
	
	Example:
		typeAttr = "foo"
		valueAttr = "bar"
		value = "mystring"
		
		would result in the following XML code:
			<foo="str" bar="mystring" />
	
	Parameters:
		value -- the value of the variable.
		node -- node to save the variable value to.
		typeAttr -- type of the variable to be saved.
		valueAttr -- value of the variable to be loaded.
	"""
	if type(value) == int:
		node.setAttribute(typeAttr, "int")
	elif type(value) == float:
		node.setAttribute(typeAttr, "float")
	elif type(value) == bool:
		node.setAttribute(typeAttr, "bool")
	elif value == None:
		node.setAttribute(typeAttr, "NoneType")
	else:
		node.setAttribute(typeAttr, "str")
		
	node.setAttribute(valueAttr, str(value))

#_____________________________________________________________________

def HandleGstPbutilsMissingMessage(message, callback, x_window_id=0):
	# Not all platforms have pbutils
	if not have_pbutils:
		return False

	#self._installing_plugins = True
	
	detail = gst.pbutils.missing_plugin_message_get_installer_detail(message)
	ctx = gst.pbutils.InstallPluginsContext()
	Globals.debug(detail)
	if x_window_id:
		ctx.set_x_id(x_window_id)
	
	ret = gst.pbutils.install_plugins_async([detail], ctx, callback)
	return True

#_____________________________________________________________________

def StringUnRepr(s):
	if sys.version > (2, 6, 0):
		import ast
		try:
			return ast.literal_eval(s)
		except ValueError, e:
			return ""
	
	if len(s) < 2:
		return ""

	quote = s[0]
	s = s[1:-1].replace("\\\\", "\\") \
	           .replace("\\t", "\t") \
	           .replace("\\n", "\n") \
	           .replace("\\r", "\r") \
	           .replace("\\" + quote, quote)
	
	strings = s.split("\\x")
	replace = strings[0]
	for string in strings[1:]:
		hex_str = string[:2]
		char = chr(int(hex_str, 16))
		replace += char + string[2:]

	return replace

#_____________________________________________________________________



#
#		THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#		THE 'COPYING' FILE FOR DETAILS
#
#		AudioBackend.py
#
#		This file offers functions to retrieve information about the
#		audio devices available with the current audio backend, and
#		the capabilities that those devices support.
#
#-------------------------------------------------------------------------------

import gst, gobject
import Globals

#=========================================================================

def ListPlaybackDevices(sink=None, probe_name=True):
	if not sink:
		# get preference from Globals
		sink = Globals.settings.playback["audiosink"]
		
	try:
		bin = gst.parse_bin_from_description(sink, False)
	except (gobject.GError, gst.ElementNotFoundError):
		Globals.debug("Cannot list playback devices: cannot parse bin", sink)
		return list()
	
	try:
		element = bin.sinks().next()
	except StopIteration:
		Globals.debug("Cannot list playback devices: no sink device in the bin", sink)
		
	return ListDeviceProbe(element, probe_name)
	
#_____________________________________________________________________
	
def ListCaptureDevices(src=None, probe_name=True):
	if not src:
		# get preference from Globals
		src = Globals.settings.recording["audiosrc"]
	try:
		bin = gst.parse_bin_from_description(src, False)
	except (gobject.GError, gst.ElementNotFoundError):
		Globals.debug("Cannot list capture devices: cannot parse bin", src)
		return list()
	
	try:
		element = bin.iterate_sources().next()
	except StopIteration:
		Globals.debug("Cannot list capture devices: no source device in the bin", src)
		
	return ListDeviceProbe(element, probe_name)
	
#_____________________________________________________________________

def ListDeviceProbe(element, probe_name):
	element_name = element.get_factory().get_property("name")
	dev_info_list = []
	
	if hasattr(element.props, "device"):
		default_device = element.__class__.props.device.default_value
		if gobject.type_is_a(element, gst.interfaces.PropertyProbe):
			element.probe_property_name("device")
			devices = element.probe_get_values_name("device")
			
			if not default_device in devices:
				dev_info_list.append((default_device, ""))
			
			if probe_name and hasattr(element.props, "device-name"):
				for dev in devices:
					element.set_property("device", dev)
					
					element.set_state(gst.STATE_PAUSED)
					# certain elements like pulsesrc won't load the device-name until STATE_PAUSED
					name = element.get_property("device-name")
					element.set_state(gst.STATE_NULL)
					
					dev_info_list.append((dev,name))
			else:
				for dev in devices:
					dev_info_list.append((dev,""))
		else:
			dev_info_list.append((default_device, ""))
	else:
		Globals.debug("Cannot list devices: property probe not supported on", element_name)
		
	return dev_info_list

#_____________________________________________________________________

def GetRecordingSampleRate(device=None):
	""" 
	Checks for available recording sample rates.
	
	Parameters:
		device -- Backend dependent device to poll for values.
	
	Returns:
		any of the following depending on the sound card:
		1) an int representing the only supported sample rate.
		2) an IntRange class with IntRange.low and IntRange.high being the min and max sample rates.
		3) a list of ints representing all the supported sample rates.
	"""
	
	src = Globals.settings.recording["audiosrc"]
	try:
		bin = gst.parse_bin_from_description(src, False)
	except (gobject.GError, gst.ElementNotFoundError):
		Globals.debug("Cannot get sample rate: cannot parse bin", src)
		return list()
	
	try:
		element = bin.iterate_sources().next()
	except StopIteration:
		Globals.debug("Cannot get sample rate: no source device in the bin", src)
		return list()
	
	if device:
		element.set_property("device", device)

	# open device (so caps are probed)
	bin.set_state(gst.STATE_PAUSED)

	try:
		pad = element.src_pads().next()
		caps = pad.get_caps()
		val = caps[0]["rate"]
	except:
		val = None
		
	# clean up
	bin.set_state(gst.STATE_NULL)
	del element, bin
	
	return val

#_____________________________________________________________________

def GetChannelsOffered(device):
	"""
	Checks for the number of available channels on a device.
	
	Parameters:
		device -- ALSA device (e.g. hw:0) to poll for available channels.
		
	Returns:
		the number of channels available on a device.
	"""
	src_desc = Globals.settings.recording["audiosrc"]
	try:
		bin = gst.parse_bin_from_description(src_desc, True)
	except (gobject.GError, gst.ElementNotFoundError):
		Globals.debug("Cannot get number of channels: cannot parse bin", src_desc)
		return 0
	
	try:
		src = bin.iterate_sources().next()
	except StopIteration:
		Globals.debug("Cannot list capture devices: no source device in the bin", src_desc)

	fakesink = gst.element_factory_make("fakesink")
	pipeline = gst.Pipeline()
	pipeline.add(bin)
	pipeline.add(fakesink)
	bin.link(fakesink)

	src.set_property("device", device)
	pipeline.set_state(gst.STATE_PAUSED)
	# block for *at most* 100ms to wait for state to change 
	pipeline.get_state(gst.SECOND / 100)
	
	try:
		#Assume the card only offers one src (we can't handle more anyway)
		for pad in src.src_pads():
			caps = pad.get_negotiated_caps()
			# in case it hasn't negotiated yet, get standard caps which are less precise.
			if not caps:
				Globals.debug("GetChannelsOffered(): Waited for STATE_PAUSED, but still no negotiated caps on", device)
				caps = pad.get_caps()
	except:
		Globals.debug("Couldn't get source pad for %s"%device)
		pipeline.set_state(gst.STATE_NULL)
		return 0
	
	nums = []
	for struct in caps:
		channels = caps[0]["channels"]
		if isinstance(channels, gst.IntRange):
			if channels.high > 20000:
				#Assume we're being given the max number of channels for gstreamer, so take low number
				nums.append(channels.low)
			else:
				#Otherwise take the high number
				nums.append(channels.high)
		else:
			nums.append(channels)

	numChannels = max(nums)
	Globals.debug("Detected channels = %s from caps:" % numChannels, caps.to_string())

	if numChannels > 8:
		# audioconvert can't handle more than 8 channels
		numChannels = 8

	#if numChannels == 2:
		#Assume one stereo input
	#	numChannels = 1

	pipeline.set_state(gst.STATE_NULL)
	return numChannels

"""
The following function, is meant for testing this file independantly from the rest.
"""
def print_device_list(l):
	for dev, name in l:
		print " " + repr(dev)
		print "  => " + repr(name)

if __name__ == "__main__":
	print "Listing ALSA capture:"
	print_device_list(ListCaptureDevices("alsasrc", True))
		
	print "Listing ALSA output:"
	print_device_list(ListPlaybackDevices("alsasink", True))
		
	print "Listing PulseAudio capture:"
	print_device_list(ListCaptureDevices("pulsesrc", True))
	
	print "Listing PulseAudio output:"
	print_device_list(ListPlaybackDevices("pulsesink", True))



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	EffectPresets.py
#	
#	This module implements support for effects presets. These presets are used
#	to store settings for single effects and multiple effects strung together
#	(called a 'chain').
#
#	The way this works is that we have a LADSPA_FACTORY_REGISTRY filled with
#	the system's LADSPA effects, LADSPA_NAME_MAP which amps LADSPA element
#	factory names (Such as ladspa-delay-5s) to the effect name (such as
#	Simple Delay) and self.effectpresetsregistry which contains a generated
#	dictionary of effects. This dictionary is search with list comprehensions
#	to get the relavent presets out.
#
#-------------------------------------------------------------------------------

import pygst
pygst.require("0.10")
import gst
import gobject
import xml.dom.minidom as xml
import os
import Globals
from Utils import *
import glob
import string

#=========================================================================

class EffectPresets(gobject.GObject):
	"""
	This class implements support for effects presets. These presets are used
	to store settings for single effects and multiple effects strung together
	(called a 'chain').

	Signals:
		"single-preset" -- The waveform date for this event has changed.
		"chain-preset" -- The starting position of this event has changed.

	"""
	__gsignals__ = {
		"single-preset" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"chain-preset" 	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () )
	}
	
	#_____________________________________________________________________	
	
	def __init__(self):
		"""
		Creates a new instance of EffectsPresets. If needed, it populates the
		LADSPA and effect presets registries.
		"""
		gobject.GObject.__init__(self)
		
		# Version of the preset files xml format
		Globals.EFFECT_PRESETS_VERSION = "0.2"
		
		"""
		This is the main dictionary of presets. It has the following structure when filled:
		
		 	effectpresetregistry[presetType][elementName][presetname][property]
		 	
		where:
			presetType = "instruments" or "effects"
			elementName = unique ladspa or instrument name (i.e. ladspa-eq or guitar)
			presetName = name of the preset (i.e. Chorus + Delay)
			property = an specific preset property (i.e. dependencies or file)
			
			*Note: all these 4 fields are dictionaries
		
		Diagram:
		
		effectpresetregistry
			|
			+--instruments
			|  |
			|  +--guitar
			|  |  |
			|  |  +--Chorus + Delay
			|  |  |  |
			|  |  |  +--instrument: guitar
			|  |  |  +--dependencies: ["effect1", "effect2"]
			|  |  |  +--file: guitar - Chorus + Delay.jpreset
			|  |  |
			|  |  +--Heavy Metal
			|  |     |
			|  |     +-- (...)
			|  |
			|  +--audiofile
			|     |
			|     +--Delay chamber
			|     |  |
			|     |  +-- (...)
			|     |
			|     +--Hum removal
			|        |
			|        +-- (...)
			|
			+--effects
			   |
			   +--ladspa-eq
			   |  |
			   |  +--Rock
			   |  |  |
			   |  |  +--dependencies: ["effect1", "effect2"]
			   |  |  +--file: ladspa-eq - Rock.jpreset
			   |  |
			   |  +--Jazz
			   |  |  |
			   |  |  +-- (...)
			   |  |
			   |  +--Pop
			   |     |
			   |     +-- (...)
			   |
			   +--ladspa-chorus
			      |
			      +--Full depth
			      |  |
			      |  +-- (...)
			      |
			      +--Bubbly dream
			         |
			         +-- (...)
		"""
		self.effectpresetregistry = {}
		
		# string used to separate the preset type from its name when generating
		# a preset filename
		self.separator = " - "
		
		# fill the different data structures with information if necessary. The LADSPA
		# structures are part of Globals.py

		if not Globals.LADSPA_NAME_MAP or not Globals.LADSPA_FACTORY_REGISTRY:
			self.FillLADSPARegistry()

		self.FillEffectsPresetsRegistry()
	
	#_____________________________________________________________________

	def SaveSingleEffect(self, label, effectdict, effectelement, effecttype):
		"""
		This method will write a single effect preset to a preset file.
		
		Parameters:
			label --the name of the effect.
			effectdict -- the effect dictionary.
			effectelement -- the effect that the user selected.
			effecttype -- the type of the effect the user selected.
		"""
		self.effectelement = effectelement
		self.effecttype = effecttype
	
		if not Globals.EFFECT_PRESETS_PATH:
			raise Exception("No preset save path specified!")
		
		doc = xml.Document()
		head = doc.createElement("JokosherPreset")
		doc.appendChild(head)
		
		head.setAttribute("version", Globals.EFFECT_PRESETS_VERSION)

		effectblock = doc.createElement("Effect")
		#effectblock.setAttribute("element", effectelement)
		#effectblock.setAttribute("effectype", effecttype)
		head.appendChild(effectblock)
		
		paramsblock = doc.createElement("Parameters")
		effectblock.appendChild(paramsblock)
		
		paramslist = ["effectelement", "effecttype"]
		
		StoreParametersToXML(self, doc, paramsblock, paramslist)
		
		settingsblock = doc.createElement("Settings")
		effectblock.appendChild(settingsblock)
		
		StoreDictionaryToXML(doc, settingsblock, effectdict)
		
		filename = self._PresetFilename(effectelement, label)
		file = open(Globals.EFFECT_PRESETS_PATH + filename, "w")
		file.write(doc.toprettyxml())
		file.close()
		
		self.emit("single-preset")
	#_____________________________________________________________________

	def SaveEffectChain(self, label, effectlist, instrumenttype):
		"""
		Write an effect chain to a preset file.
		
		Parameters:
			label -- the name of the effect.
			effectlist -- the list of effects.
			instrumenttype -- the type of instrument currently being used.
		"""		
		self.effectelement = None
		self.effecttype = None
		
		if not Globals.EFFECT_PRESETS_PATH:
			raise Exception("No effect chain preset save path specified!")
		
		doc = xml.Document()
		head = doc.createElement("JokosherPreset")
		doc.appendChild(head)
		
		head.setAttribute("version", Globals.EFFECT_PRESETS_VERSION)

		# effect chain preset files have an extra <Chain> block which mainly
		# serves to indicate which type of instrument the effect is for
		chainblock = doc.createElement("Chain")
		head.appendChild(chainblock)
			
		chaindict = {}
		chaindict["instrument"] = instrumenttype

		StoreDictionaryToXML(doc, chainblock, chaindict)

		# the structure of each <Effect> tag is not different from the single
		# effect presets, there is just an <Effect> block for each effect in
		# the chain
		for effect in effectlist:
			self.effectelement = effect["effectelement"]
			self.effecttype = effect["effecttype"]
		
			Globals.debug(self.effectelement)

			effectblock = doc.createElement("Effect")
			head.appendChild(effectblock)
						
			paramsblock = doc.createElement("Parameters")
			effectblock.appendChild(paramsblock)
			
			paramslist = ["effectelement", "effecttype"]
			
			StoreParametersToXML(self, doc, paramsblock, paramslist)
			
			settingsblock = doc.createElement("Settings")
			effectblock.appendChild(settingsblock)
			
			StoreDictionaryToXML(doc, settingsblock, effect["settings"])
		
		filename = self._PresetFilename(instrumenttype, label)
		presetfile = open(os.path.realpath(Globals.EFFECT_PRESETS_PATH + filename), "w")
		presetfile.write(doc.toprettyxml())
		presetfile.close()
		
		self.emit("chain-preset")
		
	#_____________________________________________________________________
	
	def LoadSingleEffect(self, presetName, effectelement):
		"""
		Load effect settings from a preset file for a single effect.
		
		Parameters:
			presetName -- the name of the preset to be loaded.
			effectelement -- the effect element to be loaded.
			
		Returns:
			a settings dictionary with the loaded settings for the effect or
			False if the preset file doesn't exist.
		"""
		filename = self._PresetFilename(effectelement, presetName)
		presetfile = Globals.EFFECT_PRESETS_PATH + filename
		Globals.debug(presetfile)
		
		if not os.path.exists(presetfile):
			Globals.debug("preset file does not exist")
			return False
		else:	
			xmlfile = open(presetfile, "r")
			doc = xml.parse(presetfile)

		settingstags = doc.getElementsByTagName('Effect')[0].getElementsByTagName('Settings')[0]
		settdict = LoadDictionaryFromXML(settingstags)
		
		return settdict
	
	#____________________________________________________________________
	
	def LoadEffectChain(self, presetName, instrType):
		"""
		Load settings from the preset file for an Instrument's effects chain.
		
		Parameters:
			presetName -- name of the preset to be loaded.
			
		Returns:
			a settings dictionary with the loaded settings for the effects.
		"""
		filename = self._PresetFilename(instrType, presetName)
		presetfile = Globals.EFFECT_PRESETS_PATH + filename
		
		if not os.path.exists(presetfile):
			Globals.debug("preset file does not exist")
		else:	
			xmlfile = open(presetfile, "r")
			doc = xml.parse(presetfile)

		settdict = {}
		
		for effect in doc.getElementsByTagName('Effect'):
			preftags = effect.getElementsByTagName('Parameters')[0]
			prefs = LoadDictionaryFromXML(preftags)

			settingstags = effect.getElementsByTagName('Settings')[0]
			setts = LoadDictionaryFromXML(settingstags)
			elementname = setts["name"]
			settdict[str(elementname)] = {'preferences': prefs, 'settings': setts}

		return settdict
		
	#_____________________________________________________________________
	
	def LoadSingleEffectList(self):
		"""
		TODO -- This method is not yet implemented.
		"""
		pass
		
	#_____________________________________________________________________
	
	def LoadEffectChainList(self):
		"""
		TODO -- This method is not yet implemented.
		"""
		pass
		
	#_____________________________________________________________________
	
	def DeleteSingleEffect(self, presetName, effectName):
		"""
		Removes a single effect preset.
		
		Parameters:
			presetName -- name of the preset to be removed.
			effectName -- ladspa unique name of the effect the preset 
							belongs to.
		"""
		self._DeletePresetFile(self._PresetFilename(effectName, presetName))
		self.emit("single-preset")
	
	#_____________________________________________________________________
	
	def DeleteEffectChain(self, presetName, instrType):
		"""
		Removes an effect chain preset.
		
		Parameters:
			presetName -- name of the preset to be removed.
			instrType -- type of the Instrument the preset belongs to.
		"""
		self._DeletePresetFile(self._PresetFilename(instrType, presetName))
		self.emit("chain-preset")
	
	#_____________________________________________________________________
	
	def _DeletePresetFile(self, filename):
		"""
		Removes a preset file.
		
		Parameters:
			filename -- name of the preset file to remove.
		"""
		presetFile = os.path.expanduser(Globals.EFFECT_PRESETS_PATH + filename)
		
		if os.path.isfile(presetFile):
			os.remove(presetFile)
	
	#_____________________________________________________________________
	
	def _PresetFilename(self, prefix, name):
		"""
		Creates the correct preset filename according to the parameters.
		
		Examples:	
			PresetFilename("Guitar", "Soloist") will output:
				"/Guitar %separator% Soloist.jpreset"
				
			PresetFilename("ladspa-delay", "5ms deep delay") will output:
				"/ladspa-delay %separator% 5ms deep delay.jpreset"
			
			where %separator% is the separator string defined inside __init__
			
		Parameters:
			prefix -- unique ladspa shortname or instrType.
			name -- name of the preset.
			
		Returns:
			a properly formatted preset filename string.
		"""
		return ("/%s%s%s.jpreset") % (prefix, self.separator, name)
	
	#_____________________________________________________________________
	
	def FillEffectsPresetsRegistry(self):
		"""
		Load all chain/effect presets into the main presets registry.
		"""
		Globals.debug("\tReading in presets...")
		presetsfiles = glob.glob(Globals.EFFECT_PRESETS_PATH + "/*.jpreset")
		
		self.effectpresetregistry = {}
		self.effectpresetregistry["instruments"] = {}
		self.effectpresetregistry["effects"] = {}
		
		for file_ in presetsfiles:
			preset = {}
			depslist = []
			presetname = None
			effectName = None
			
			if not os.path.exists(file_):
				Globals.debug("preset file does not exist")
			else:	
				xmlfile = open(file_, "r")
				doc = xml.parse(file_)

			# True if the loaded preset corresponds to an effect chain, False otherwise
			isChain = None
			
			try:	
				instrument = doc.getElementsByTagName('Chain')[0].getElementsByTagName('instrument')[0].getAttribute('value')
				isChain = True
			except:
				instrument = None
					
			for effect in doc.getElementsByTagName("Effect"):
				paramtags = effect.getElementsByTagName("Parameters")[0]

				for node in paramtags.childNodes:
					if node.nodeType == xml.Node.ELEMENT_NODE:
						if node.getAttribute("type") == "int":
							pass
						elif node.getAttribute("type") == "float":
							pass
						else:
							if node.tagName == "effectelement":
								depslist.append(str(node.getAttribute("value")))
								effectName = str(node.getAttribute("value"))
			
			presetname = file_.replace(str(Globals.EFFECT_PRESETS_PATH + "/"), "")
			presetfile = presetname
			
			# extract the preset name from the prefix
			presetname = presetname.split(self.separator, 1)
			if len(presetname) == 1:
				# the filename doesn't have a prefix. Could be an old or non-compliant file
				# TODO: should upgrade the filename or it won't load
				presetname = presetname[0]
			else:
				presetname = presetname[1]
			
			presetname = presetname.replace(".jpreset", "")	
			preset["dependencies"] = set(depslist)
			preset["file"] = str(presetfile)
			
			if isChain:
				preset["instrument"] = str(instrument)
				presetType = "instruments"
				elementName = instrument
			else:
				presetType = "effects"
				elementName = effectName
				
			# create the elementName dir if it doesn't exist
			try:
				self.effectpresetregistry[presetType][elementName]
			except KeyError:
				self.effectpresetregistry[presetType][elementName] = {}
			
			self.effectpresetregistry[presetType][elementName][presetname] = preset
			
		Globals.debug("\t...done.")
		
	#_____________________________________________________________________
	
	def FillLADSPARegistry(self):
		"""
		Fill Globals.LADSPA_FACTORY_REGISTRY with effects on the system. This
		is to ensure that only presets with effects on the current system are listed.
		"""
		Globals.debug("Filling LADSPA Registry")
		
		##make sure all the structures are empty before we append to them
		Globals.LADSPA_NAME_MAP=[]
		Globals.LADSPA_FACTORY_REGISTRY = None
		effects = []

		ladspaFactoryList = gst.registry_get_default().get_feature_list_by_plugin("ladspa")
		
		for factory in ladspaFactoryList:
			if isinstance(factory, gst.ElementFactory):
				# from the list of LADSPA effects we check which ones only
				# have a single sink and a single src so we know they work
				if factory.get_num_pad_templates() == 2:
					pads = factory.get_static_pad_templates()
					sinkpads = len( [pad for pad in pads if pad.direction == gst.PAD_SINK] )
					srcpads = len( [pad for pad in pads if pad.direction == gst.PAD_SRC] )
					
					if srcpads == 1 and sinkpads == 1:
						effects.append(factory.get_name())
						Globals.LADSPA_NAME_MAP.append((factory.get_name(), factory.get_longname()))

		Globals.debug("\t", len(effects), "LADSPA effects loaded")
		Globals.LADSPA_FACTORY_REGISTRY = set(effects)
		
	#_____________________________________________________________________
	
#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	NewDialog.py
#	
#	This module shows the New Project which is used to create a new Jokosher
#	project.
#
#-------------------------------------------------------------------------------

import gtk.glade
import os
import ProjectManager
import PlatformUtils
import Globals
import gettext
_ = gettext.gettext
import ProjectTemplate
import ProjectTemplateDialog

class NewProjectDialog:
	"""
	This module shows the New Project which is used to create a new Jokosher project.
	"""
	#_____________________________________________________________________	
	
	def __init__(self, parent):
		"""
		Creates a new instance of NewProjectDialog.
		
		The dialog is used to create a new Project based on the user's input.
		
		Parameters:
			parent -- reference to the MainApp Jokosher window.
		"""
		self.parent = parent
		
		self.res = gtk.glade.XML(Globals.GLADE_PATH, "NewProjectDialog")

		self.signals = {
			"on_OK_clicked" : self.OnOK,
			"on_Cancel_clicked" : self.OnCancel,
			"on_editbutton_clicked" : self.OnEdit,
			"on_Template_toggled" : self.TemplateCheckboxToggled
		}
		
		self.res.signal_autoconnect(self.signals)
		
		self.dlg = self.res.get_widget("NewProjectDialog")

		self.sideimage = self.res.get_widget("sideimage")
		self.sideimage.set_from_file(os.path.join(Globals.IMAGE_PATH, "newproject.png"))
		
		self.name = self.res.get_widget("name")
		self.name.set_activates_default(True)
		self.author = self.res.get_widget("author")
		self.author.set_activates_default(True)
		
		self.template = ProjectTemplate.ProjectTemplate()
		self.template_checkbox = self.res.get_widget("template_checkbox")
		self.templatehbox = self.res.get_widget("template_hbox")
		self.templatehbox.set_sensitive(False)
		self.templatecombo = self.res.get_widget("template_combo")
		self.templatecombo.clear()
		
		self.templatemodel = gtk.ListStore(str)
		for files in self.template.LoadDictionaryOfInstrumentsFromTemplateFile().iterkeys():
			self.templatemodel.append((files,))
		self.templatecombo.set_model(self.templatemodel)
		
		text = gtk.CellRendererText()
		self.templatecombo.pack_start(text)
		self.templatecombo.add_attribute(text, "text", 0)
		
		self.templatecombo.set_active(0)

		# Default author to name of currently logged in user
		self.author.set_text(PlatformUtils.getFullName())
		
		self.okbutton = self.res.get_widget("okButton")
		self.okbutton.set_flags(gtk.CAN_DEFAULT)
		self.okbutton.grab_default()
		
		self.folder = gtk.FileChooserButton(_("Select A Folder"))
		self.folder.set_local_only(True)
		self.folder.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
		self.res.get_widget("table5").attach(self.folder, 1, 2, 1, 2)
		self.folder.show()
		
		# Set the default folder of 'folder' (a FileChooserButton)
		if os.path.exists(Globals.settings.general["projectfolder"]):
			self.folder.set_current_folder(Globals.settings.general["projectfolder"])
		else:
			self.folder.set_current_folder(os.path.expanduser("~"))

		self.dlg.resize(350, 300)
		self.dlg.set_icon(self.parent.icon)
		self.dlg.set_transient_for(self.parent.window)

	#_____________________________________________________________________
	
	def TemplateCheckboxToggled(self, widget):
		"""
		Called when the template checkbox is activated.
		The template hbox becomes inactive if the template checkbox is activated.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly. 
		"""
		self.templatehbox.set_sensitive(self.template_checkbox.get_active())
		
	#_____________________________________________________________________	
		
	def OnOK(self, button):
		"""
		Tries to create and set a new Project with the user input name, author
		and location.
		If the process fails, a message is issued to the user stating the error.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		name = self.name.get_text()
		if not name:
			name = _("New Project")
			
		author = self.author.get_text()
		if not author:
			author = _("Unknown Author")
			
		# CreateNewProject expects a URI	
		folder = PlatformUtils.pathname2url(self.folder.get_filename())
		
		# Save the selected folder as the default folder
		Globals.settings.general["projectfolder"] = folder
		Globals.settings.write()
		if not folder:
			folder = "~"

		try:
			project = ProjectManager.CreateNewProject(folder, name, author)
		except ProjectManager.CreateProjectError, e:
			if e.errno == 1:
				message = _("Could not initialize project.")
			elif e.errno == 2:
				message = _("A file or folder with this name already exists. Please choose a different project name and try again.")
			elif e.errno == 3:
				message = _("The file or folder location is write-protected.")
			elif e.errno == 4:
				message = _("Invalid name or author.")
			elif e.errno == 5:
				message = _("The URI scheme given is either invalid or not supported")
			elif e.errno == 6:
				message = "%s %s" % (_("Unable to load required Gstreamer plugin:"), e.message)
			
			# show the error dialog with the relavent error message	
			dlg = gtk.MessageDialog(self.dlg,
				gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
				gtk.MESSAGE_ERROR,
				gtk.BUTTONS_OK,
				_("Unable to create project.\n\n%s") % message)
			dlg.run()
			dlg.destroy()
		else:
			self.AddTemplateToProject(project)
			self.parent.SetProject(project)
			self.dlg.destroy()
		
	#_____________________________________________________________________	
	
	def OnCancel(self, button):
		"""
		Destroys the dialog when the cancel button is pressed.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.dlg.destroy()

	#_____________________________________________________________________
	
	def OnEdit(self, button):
		"""
		Displays the ProjectTemplate dialog, allowing the user to create, delete and modify project
		templates.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		ProjectTemplateDialog.ProjectTemplateDialog(self, self.template)
		
	#_____________________________________________________________________
	
	def ReturnProjectInstrumentTuples(self, name):
		"""
		Ths method will return a tuple containing the name, type and pixbuf of the instruments in a given template file.
		
		Parameters:
			name -- the name of the template to use.
		
		Returns:
			instruments -- a list of instruments containing tuples required for project.AddInstruments()
		"""
		return self.template.LoadDictionaryOfInstrumentsFromTemplateFile()[name]
		
	#_____________________________________________________________________
	
	def AddTemplateToProject(self, project):
		"""
		Adds template instruments to the user's project.
		"""
		if self.template_checkbox.get_active():
			active = self.templatecombo.get_model()[self.templatecombo.get_active()][0]
			tuples = self.ReturnProjectInstrumentTuples(active)
			instrList = []
			for name, type, pixbuf in tuples:
				instrList.append( (name, type))
			project.AddInstruments(instrList)
	
	#_____________________________________________________________________
	
#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Project.py
#	
#	This module is the central non-gui class for Jokosher. It saves and loads
#	project files, and handles any project wide functionality including;
#	settings, instruments, recording, playing, exporint, zooming, scrolling, 
#	undo, redo, volume, etc.
#
#-------------------------------------------------------------------------------

import pygst
pygst.require("0.10")
import gst
import gobject
import os, os.path
import gzip
import re

import TransportManager
import UndoSystem, IncrementalSave
import Globals
import xml.dom.minidom as xml
import Instrument, Event
import Utils
import AudioBackend
import ProjectManager
import PlatformUtils

#=========================================================================

class Project(gobject.GObject):
	"""
	This class maintains all of the information required about single Project. It also
	saves and loads Project files.
	"""
	
	""" The Project structure version. Will be useful for handling old save files. """
	Globals.VERSION = "0.11.1"
	
	""" The audio playback state enum values """
	AUDIO_STOPPED, AUDIO_RECORDING, AUDIO_PLAYING, AUDIO_PAUSED, AUDIO_EXPORTING = range(5)
	
	""" String constants for incremental save """
	INCREMENTAL_SAVE_EXT = ".incremental"
	INCREMENTAL_SAVE_DELIMITER = "\n<<delimiter>>\n"
	
	"""
	Signals:
		"audio-state" -- The status of the audio system has changed. See below:
			"audio-state::play" -- The audio started playing.
			"audio-state::pause" -- The audio is paused.
			"audio-state::record" -- The audio started recording.
			"audio-state::stop" -- The playback or recording was stopped.
			"audio-state::export-start" -- The audio is being played to a file.
			"audio-state::export-stop" -- The export to a file has completed.
		"bpm" -- The beats per minute value was changed.
		"click-track" -- The volume of the click track changed.
		"gst-bus-error" -- An error message was posted to the pipeline. Two strings are also send with the error details.
		"incremental-save" -- An action was logged to the .incremental file.
		"instrument" -- The instruments for this project have changed. The instrument instance will be passed as a parameter. See below:
			"instrument::added" -- An instrument was added to this project.
			"instrument::removed" -- An instrument was removed from this project.
			"instrument::reordered" -- The order of the instruments for this project changed.
		"time-signature" -- The time signature values were changed.
		"undo" -- The undo or redo stacks for this project have been changed.
		"view-start" -- The starting position of the view of this project's timeline has changed.
		"volume" -- This master volume value for this project has changed.
		"zoom" -- The zoom level of this project's timeline has changed.
	"""
	
	__gsignals__ = {
		"audio-state"		: ( gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_DETAILED, gobject.TYPE_NONE, () ),
		"bpm"			: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"click-track"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_DOUBLE,) ),
		"gst-bus-error"	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING, gobject.TYPE_STRING) ),
		"incremental-save" : ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"instrument"		: ( gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_DETAILED, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,) ),
		"time-signature"	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"undo"			: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"view-start"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"volume"			: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"zoom"			: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () )
	}

	#_____________________________________________________________________

	def __init__(self):
		"""
		Creates a new instance of Project with default values.
		"""
		gobject.GObject.__init__(self)
		
		self.author = ""			#user specified author of this project
		self.name = ""				#the name of this project
		self.notes = ""				#user specified notes for the project
		self.projectfile = ""		#the name of the project file, complete with path
		self.audio_path = ""
		self.levels_path = ""
		self.___id_list = []		#the list of IDs that have already been used, to avoid collisions
		self.instruments = []		#the list of instruments held by this project
		self.graveyard = []			# The place where deleted instruments are kept, to later be retrieved by undo functions
		#used to delete copied audio files if the event that uses them is not saved in the project file
		#also contains paths to levels_data files corresponding to those audio files
		self.deleteOnCloseAudioFiles = []	# WARNING: any paths in this list will be deleted on exit!
		self.clipboardList = []		#The list containing the events to cut/copy
		self.viewScale = 25.0		#View scale as pixels per second
		self.viewStart= 0.0			#View offset in seconds
		self.soloInstrCount = 0		#number of solo instruments (to know if others must be muted)
		self.audioState = self.AUDIO_STOPPED	#which audio state we are currently in
		self.exportPending = False	# True if we are waiting to start an export
		self.exportFilename = ""
		self.bpm = 120
		self.meter_nom = 4		# time signature numerator
		self.meter_denom = 4		# time signature denominator
		self.clickbpm = 120			#the number of beats per minute that the click track will play
		self.clickVolumeValue = 0	#The value of the click track volume between 0.0 and 1.0
		#Keys are instruments which are recording; values are 3-tuples of the event being recorded, the recording bin and bus handler id
		self.recordingEvents = {}	#Dict containing recording information for each recording instrument
		self.volume = 1.0			#The volume setting for the entire project
		self.level = 0.0			#The level of the entire project as reported by the gstreamer element
		self.currentSinkString = None	#to keep track if the sink changes or not

		self.hasDoneIncrementalSave = False	# True if we have already written to the .incremental file from this project.
		self.isDoingIncrementalRestore = False # If we are currently restoring incremental save actions

		# Variables for the undo/redo command system
		self.__unsavedChanges = False	#This boolean is to indicate if something which is not on the undo/redo stack needs to be saved
		self.__undoStack = []			#not yet saved undo commands
		self.__redoStack = []			#not yet saved actions that we're undone
		self.__savedUndoStack = []		#undo commands that have already been saved in the project file
		self.__savedRedoStack = []		#redo commands that have already been saved in the project file
		self.__performingUndo = False	#True if we are currently in the process of performing an undo command
		self.__performingRedo = False	#True if we are currently in the process of performing a redo command
		self.__savedUndo = False		#True if we are performing an undo/redo command that was previously saved
	
		
		# CREATE GSTREAMER ELEMENTS AND SET PROPERTIES #
		self.mainpipeline = gst.Pipeline("timeline")
		self.playbackbin = gst.Bin("playbackbin")
		self.adder = gst.element_factory_make("adder")
		self.postAdderConvert = gst.element_factory_make("audioconvert")
		self.masterSink = self.MakeProjectSink()
		
		self.levelElement = gst.element_factory_make("level", "MasterLevel")
		self.levelElement.set_property("interval", gst.SECOND / 50)
		self.levelElement.set_property("message", True)
		
		#Restrict adder's output caps due to adder bug 341431
		self.levelElementCaps = gst.element_factory_make("capsfilter", "levelcaps")
		capsString = "audio/x-raw-float,rate=44100,channels=2,width=32,endianness=1234"
		caps = gst.caps_from_string(capsString)
		self.levelElementCaps.set_property("caps", caps)
		
		# ADD ELEMENTS TO THE PIPELINE AND/OR THEIR BINS #
		self.mainpipeline.add(self.playbackbin)
		Globals.debug("added project playback bin to the pipeline")
		for element in [self.adder, self.levelElementCaps, self.postAdderConvert, self.levelElement, self.masterSink]:
			self.playbackbin.add(element)
			Globals.debug("added %s to project playbackbin" % element.get_name())

		# LINK GSTREAMER ELEMENTS #
		self.adder.link(self.levelElementCaps)
		self.levelElementCaps.link(self.postAdderConvert)
		self.postAdderConvert.link(self.levelElement)
		self.levelElement.link(self.masterSink)
		
		# CONSTRUCT CLICK TRACK BIN #
		self.clickTrackBin = gst.Bin("Click_Track_Bin")
		self.clickTrackAudioSrc = gst.element_factory_make("audiotestsrc", "Click_Track_AudioSource")
		self.clickTrackAudioSrc.set_property("wave", 3)
		self.clickTrackVolume = gst.element_factory_make("volume", "Click_Track_Volume")
		self.clickTrackVolume.set_property("mute", True)
		self.clickTrackConvert = gst.element_factory_make("audioconvert", "Click_Track_Audioconvert")
		
		self.playbackbin.add(self.clickTrackBin)
		for element in [self.clickTrackAudioSrc, self.clickTrackVolume, self.clickTrackConvert]:
			self.clickTrackBin.add(element)
		
		clickTrackSrc = gst.GhostPad("src", self.clickTrackConvert.get_pad("src"))
		self.clickTrackBin.add_pad(clickTrackSrc)
		self.clickTrackController = gst.Controller(self.clickTrackAudioSrc, "volume")
		
		self.clickTrackAudioSrc.link(self.clickTrackVolume)
		self.clickTrackVolume.link(self.clickTrackConvert)
		self.clickTrackBin.link(self.adder)
		# /END OF GSTREAMER BITS #
		
		# set up the bus message callbacks
		self.bus = self.mainpipeline.get_bus()
		self.bus.add_signal_watch()
		self.Mhandler = self.bus.connect("message::element", self.__PipelineBusLevelCb)
		self.EOShandler = self.bus.connect("message::eos", self.Stop)
		self.Errorhandler = self.bus.connect("message::error", self.__PipelineBusErrorCb)
		
		#initialize the transport mode
		self.transportMode = TransportManager.TransportManager.MODE_BARS_BEATS
		self.transport = TransportManager.TransportManager(self.transportMode, self)

		self.PrepareClick()
	
	#_____________________________________________________________________
	
	def Play(self, newAudioState=None):
		"""
		Start playback or recording.
		
		Parameters:
			newAudioState -- determines the Project audio state to set when playback commences:
							AUDIO_PAUSED or AUDIO_PLAYING = move the graphical indicator along playback.
							AUDIO_EXPORTING = perform playback without moving the graphical bar.
			recording -- determines if the Project should only playback or playback and record:
						True = playback and record.
						False = playback only.
		"""
		if not newAudioState:
			newAudioState = self.AUDIO_PLAYING
		
		Globals.debug("play() in Project.py")
		Globals.debug("current state:", self.mainpipeline.get_state(0)[1].value_name)

		for ins in self.instruments:
			ins.PrepareController()
		
		if not self.GetIsPlaying():
			# Connect the state changed handler
			self.state_id = self.bus.connect("message::state-changed", self.__PlaybackStateChangedCb, newAudioState)
			#set to PAUSED so the transport manager can seek first (if needed)
			#the pipeline will be set to PLAY by self.state_changed()
			self.mainpipeline.set_state(gst.STATE_PAUSED)
		else:
			# we are already paused or playing, so just start the transport manager
			self.transport.Play(newAudioState)
		
		Globals.debug("just set state to PAUSED")
		
	#_____________________________________________________________________

	def Pause(self):
		self.transport.Pause()
	
	#_____________________________________________________________________
	
	def Stop(self, bus=None, message=None):
		"""
		Stop playback or recording
		
		Parameters:
			bus -- reserved for GStreamer callbacks, don't use it explicitly.
			message -- reserved for GStreamer callbacks, don't use it explicitly.
		"""

		Globals.debug("Stop pressed, about to set state to READY")
		Globals.debug("current state:", self.mainpipeline.get_state(0)[1].value_name)
		
		#If we've been recording then add new events to instruments
		for instr, (event, bin, handle) in self.recordingEvents.iteritems():
			instr.FinalizeRecording(event)
			self.bus.disconnect(handle)

		self.TerminateRecording()
		
	#_____________________________________________________________________

	def TerminateRecording(self):
		"""
		Terminate all instruments. Disregards recording when an 
		error occurs after instruments have started.
		"""
		Globals.debug("Terminating recording.")
		self.transport.Stop()
		Globals.debug("State just set to READY")
		
		#Relink instruments and stop their recording bins
		for instr, (event, bin, handle) in self.recordingEvents.iteritems():
			try:
				Globals.debug("Removing recordingEvents bin")
				self.mainpipeline.remove(bin)
			except:
				pass #Already removed from another instrument
			Globals.debug("set state to NULL")
			bin.set_state(gst.STATE_NULL)
			instr.AddAndLinkPlaybackbin()

		self.recordingEvents = {}

	#_____________________________________________________________________
	
	def Record(self):
		"""
		Start recording all selected instruments.
		"""

		Globals.debug("pre-record state:", self.mainpipeline.get_state(0)[1].value_name)
		
		#Add all instruments to the pipeline
		self.recordingEvents = {}
		devices = {}
		capture_devices = AudioBackend.ListCaptureDevices(probe_name=False)
		if not capture_devices:
			capture_devices = ((None,None),)
		
		default_device = capture_devices[0][0]
		
		for device, deviceName in capture_devices:
			devices[device] = []
			for instr in self.instruments:
				if instr.isArmed and (instr.input == device or device is None):
					instr.RemoveAndUnlinkPlaybackbin()
					devices[device].append(instr)
				elif instr.isArmed and instr.input is None:
					instr.RemoveAndUnlinkPlaybackbin()
					devices[default_device].append(instr)
		

		for device, recInstruments in devices.iteritems():
			if len(recInstruments) == 0:
				#Nothing to record on this device
				continue

			if device is None:
				# assume we are using a backend like JACK which does not allow
				#us to do device selection.
				channelsNeeded = len(recInstruments)
			else:
				channelsNeeded = AudioBackend.GetChannelsOffered(device)

			
			if channelsNeeded > 1: #We're recording from a multi-input device
				recordingbin = gst.Bin("recording bin")
				recordString = Globals.settings.recording["audiosrc"]
				srcBin = gst.parse_bin_from_description(recordString, True)
				try:
					src_element = recordingbin.iterate_sources().next()
				except StopIteration:
					pass
				else:
					if hasattr(src_element.props, "device"):
						src_element.set_property("device", device)
				
				caps = gst.caps_from_string("audio/x-raw-int;audio/x-raw-float")

				sampleRate = Globals.settings.recording["samplerate"]
				try:
					sampleRate = int(sampleRate)
				except ValueError:
					sampleRate = 0
				# 0 means for "autodetect", or more technically "don't use any rate caps".
				if sampleRate > 0:
					for struct in caps:
						struct.set_value("rate", sampleRate)

				for struct in caps:
					struct.set_value("channels", channelsNeeded)

				Globals.debug("recording with capsfilter:", caps.to_string())
				capsfilter = gst.element_factory_make("capsfilter")
				capsfilter.set_property("caps", caps)
				
				split = gst.element_factory_make("deinterleave")
				convert = gst.element_factory_make("audioconvert")
				
				recordingbin.add(srcBin, split, convert, capsfilter)
				
				srcBin.link(convert) 
				convert.link(capsfilter)
				capsfilter.link(split)
				
				split.connect("pad-added", self.__RecordingPadAddedCb, recInstruments, recordingbin)
				Globals.debug("Recording in multi-input mode")
				Globals.debug("adding recordingbin")
				self.mainpipeline.add(recordingbin)
			else:
				instr = recInstruments[0]
				event = instr.GetRecordingEvent()
				
				encodeString = Globals.settings.recording["fileformat"]
				recordString = Globals.settings.recording["audiosrc"]
				
				sampleRate = 0
				try:
					sampleRate = int( Globals.settings.recording["samplerate"] )
				except ValueError:
					pass
				# 0 means for "autodetect", or more technically "don't use any caps".
				if sampleRate > 0:
					capsString = "audio/x-raw-int,rate=%s ! audioconvert" % sampleRate
				else:
					capsString = "audioconvert"
					
				# TODO: get rid of this entire string; do it manually
				pipe = "%s ! %s ! level name=recordlevel ! audioconvert ! %s ! filesink name=sink"
				pipe %= (recordString, capsString, encodeString)
				
				Globals.debug("Using pipeline: %s" % pipe)
				
				recordingbin = gst.parse_bin_from_description(pipe, False)
				
				filesink = recordingbin.get_by_name("sink")
				level = recordingbin.get_by_name("recordlevel")
				
				filesink.set_property("location", event.GetAbsFile())
				level.set_property("interval", int(event.LEVEL_INTERVAL * gst.SECOND))
				
				#update the levels in real time
				handle = self.bus.connect("message::element", event.recording_bus_level)
				
				try:
					src_element = recordingbin.iterate_sources().next()
				except StopIteration:
					pass
				else:
					if hasattr(src_element.props, "device"):
						src_element.set_property("device", device)
				
				self.recordingEvents[instr] = (event, recordingbin, handle)
				
				Globals.debug("Recording in single-input mode")
				Globals.debug("Using input track: %s" % instr.inTrack)
				
				Globals.debug("adding recordingbin")
				self.mainpipeline.add(recordingbin)
		
		#start the pipeline!
		self.Play(newAudioState=self.AUDIO_RECORDING)
		
	#_____________________________________________________________________

	def Export(self, filename, encodeBin):
		"""
		Export to location filename with format specified by format variable.
		
		Parameters:
			filename -- filename where the exported audio will be saved.
			encodeBin -- the gst-launch syntax string of the encoder as used in Globals.EXPORT_FORMATS:
					for ogg: "vorbisenc ! oggmux"
					for mp3: "lame"
					for wav: "wavenc"
		"""
		#try to create encoder/muxer first, before modifying the main pipeline.
		try:
			self.encodebin = gst.parse_bin_from_description(encodeBin, True)
		except gobject.GError, e:
			if e.code == gst.PARSE_ERROR_NO_SUCH_ELEMENT:
				error_no = ProjectManager.ProjectExportException.MISSING_ELEMENT
			else:
				error_no = ProjectManager.ProjectExportException.INVALID_ENCODE_BIN
			raise ProjectManager.ProjectExportException(error_no, e.message)

		#stop playback because some elements will be removed from the pipeline
		self.Stop()
		
		#remove and unlink the alsasink
		self.playbackbin.remove(self.masterSink, self.levelElement)
		self.postAdderConvert.unlink(self.levelElement)
		self.levelElement.unlink(self.masterSink)
		
		#create filesink
		self.outfile = gst.element_factory_make("filesink", "export_file")
		self.outfile.set_property("location", filename)
		self.playbackbin.add(self.outfile)

		self.playbackbin.add(self.encodebin)
		self.postAdderConvert.link(self.encodebin)
		self.encodebin.link(self.outfile)
			
		#disconnect the bus message handler so the levels don't change
		self.bus.disconnect(self.Mhandler)
		self.bus.disconnect(self.EOShandler)
		self.EOShandler = self.bus.connect("message::eos", self.TerminateExport)
		
		self.exportPending = True
		self.exportFilename = filename
		#start the pipeline!
		self.Play(newAudioState=self.AUDIO_EXPORTING)
		self.emit("audio-state::export-start")

	#_____________________________________________________________________
	
	def TerminateExport(self, bus=None, message=None):
		"""
		GStreamer End Of Stream handler. It is connected to eos on 
		mainpipeline while export is taking place.
		
		Parameters:
			bus -- reserved for GStreamer callbacks, don't use it explicitly.
			message -- reserved for GStreamer callbacks, don't use it explicitly.
		"""
		
		if not self.GetIsExporting():
			return
	
		#stop playback because some elements will be removed from the pipeline
		self.Stop()
	
		self.bus.disconnect(self.EOShandler)
		self.Mhandler = self.bus.connect("message::element", self.__PipelineBusLevelCb)
		self.EOShandler = self.bus.connect("message::eos", self.Stop)
		
		#remove the filesink and encoder
		self.playbackbin.remove(self.outfile, self.encodebin)		
		self.postAdderConvert.unlink(self.encodebin)
			
		#dispose of the elements
		self.outfile.set_state(gst.STATE_NULL)
		self.encodebin.set_state(gst.STATE_NULL)
		del self.outfile, self.encodebin
		
		#re-add all the alsa playback elements
		self.playbackbin.add(self.masterSink, self.levelElement)
		self.postAdderConvert.link(self.levelElement)
		self.levelElement.link(self.masterSink)
		
		self.emit("audio-state::export-stop")
	
	#_____________________________________________________________________
	
	def GetExportProgress(self):
		"""
		Returns a tuple with the number of seconds exported
		and the number of total seconds.
		"""
		if self.exportPending or self.GetIsExporting():
			try:
				#total = self.mainpipeline.query_duration(gst.FORMAT_TIME)[0]
				total = self.GetProjectLength() * gst.SECOND
				current = self.mainpipeline.query_position(gst.FORMAT_TIME)[0]
			except gst.QueryError:
				return (-1, -1)
			else:
				if current > total:
					total = current
					self.TerminateExport()
				return (float(current)/gst.SECOND, float(total)/gst.SECOND)
		else:
			return (100, 100)
		
	#_____________________________________________________________________
	
	def GetIsPlaying(self):
		"""
		Returns true if the Project is not in the stopped state,
		because paused, playing and recording are all forms of playing.
		"""
		return self.audioState != self.AUDIO_STOPPED
	
	#_____________________________________________________________________
	
	def GetIsRecording(self):
		"""
		Returns true if the Project is in the recording state.
		"""
		return self.audioState == self.AUDIO_RECORDING
	
	#_____________________________________________________________________
	
	def GetIsExporting(self):
		"""
		Returns true if the Project is not in the stopped state,
		because paused, playing and recording are all forms of playing.
		"""
		return self.audioState == self.AUDIO_EXPORTING
	
	#_____________________________________________________________________
	
	def SetAudioState(self, newState):
		"""
		Set the Project's audio state to the new state enum value.
		
		Parameters:
			newState -- the new state to set the Project to.
		"""
		self.audioState = newState
		if newState == self.AUDIO_PAUSED:
			self.emit("audio-state::pause")
		elif newState == self.AUDIO_PLAYING:
			self.emit("audio-state::play")
		elif newState == self.AUDIO_STOPPED:
			self.emit("audio-state::stop")
		elif newState == self.AUDIO_RECORDING:
			self.emit("audio-state::record")
		elif newState == self.AUDIO_EXPORTING:
			self.exportPending = False
		
	#_____________________________________________________________________
	
	def __RecordingPadAddedCb(self, elem, pad, recInstruments, bin):
		"""
		Handles new pad creation on the GStreamer channel split element
		when recording multiple streams at once. This method will be called
		when the new pad is ready to be connected to the encoder.
		
		Parameters:
			elem -- the GStreamer channel split element.
			pad -- the new src pad on the channel split element.
			recInstruments -- list with all Instruments currently recording.
			bin -- the bin that stores all the recording elements.
		"""
		# SRC template: 'src%d'
		padname = pad.get_name()
		try:
			index = int(padname[3:])
		except ValueError:
			Globals.debug("Cannot start multichannel record: pad name does not match 'src%d':", padname)
			return

		for instr in recInstruments:
			if instr.inTrack == index:
				event = instr.GetRecordingEvent()
				
				# TODO: get rid of string concatentation
				encodeString = Globals.settings.recording["fileformat"]
				pipe = "queue ! audioconvert ! level name=recordlevel ! audioconvert ! %s ! filesink name=sink"
				pipe %= encodeString
				
				encodeBin = gst.parse_bin_from_description(pipe, True)
				bin.add(encodeBin)
				pad.link(encodeBin.get_pad("sink"))
				
				filesink = bin.get_by_name("sink")
				level = bin.get_by_name("recordlevel")
				
				filesink.set_property("location", event.GetAbsFile())
				level.set_property("interval", int(event.LEVEL_INTERVAL * gst.SECOND))
				
				handle = self.bus.connect("message::element", event.recording_bus_level)
				
				# since we are adding the encodebin to an already playing pipeline, sync up there states
				encodeBin.set_state(gst.STATE_PLAYING)

				self.recordingEvents[instr] = (event, bin, handle)
				Globals.debug("Linked recording channel: instrument (%s), track %d" % (instr.name, instr.inTrack))
				break

	#_____________________________________________________________________
	
	def __PlaybackStateChangedCb(self, bus, message, newAudioState):
		"""
		Handles GStreamer statechange events when the pipline is changing from
		STATE_READY to STATE_PAUSED. Once STATE_PAUSED has been reached, this
		function will tell the transport manager to start playing.
		
		Parameters:
			bus -- reserved for GStreamer callbacks, don't use it explicitly.
			message -- reserved for GStreamer callbacks, don't use it explicitly.
			newAudioState -- the new Project audio state the transport manager
							should set when playback starts.
		"""
		Globals.debug("STATE CHANGED")
		change_status, new, pending = self.mainpipeline.get_state(0)
		Globals.debug("-- status:", change_status.value_name)
		Globals.debug("-- pending:", pending.value_name)
		Globals.debug("-- new:", new.value_name)

		#Move forward to playing when we reach paused (check pending to make sure this is the final destination)
		if new == gst.STATE_PAUSED and pending == gst.STATE_VOID_PENDING and not self.GetIsPlaying():
			bus.disconnect(self.state_id)
			#The transport manager will seek if necessary, and then set the pipeline to STATE_PLAYING
			self.transport.Play(newAudioState)

	#_____________________________________________________________________
	
	def __PipelineBusLevelCb(self, bus, message):
		"""
		Handles GStreamer bus messages about the currently reported level
		for the Project or any of the Instruments.
		
		Parameters:
			bus -- reserved for GStreamer callbacks, don't use it explicitly.
			message -- reserved for GStreamer callbacks, don't use it explicitly.
			
		Returns:
			True -- TODO
		"""
		struct = message.structure
		
		if struct and struct.get_name() == "level":
			if not message.src is self.levelElement:
				for instr in self.instruments:
					if message.src is instr.levelElement:
						instr.SetLevel(Utils.DbToFloat(struct["decay"][0]))
						break
			else:
				self.SetLevel(Utils.DbToFloat(struct["decay"][0]))
			
		return True

	#_____________________________________________________________________

	def __PipelineBusErrorCb(self, bus, message):
		"""
		Handles GStreamer error messages.
		
		Parameters:
			bus -- reserved for GStreamer callbacks, don't use it explicitly.
			message -- reserved for GStreamer callbacks, don't use it explicitly.
		"""
		error, debug = message.parse_error()
		
		Globals.debug("Gstreamer bus error:", str(error), str(debug))
		Globals.debug("Domain: %s, Code: %s" % (error.domain, error.code))
		Globals.debug("Message:", error.message)
		
		if error.domain == gst.STREAM_ERROR and Globals.DEBUG_GST:
			self.DumpDotFile()
		
		self.emit("gst-bus-error", str(error), str(debug))

	#_____________________________________________________________________
	
	def DumpDotFile(self):
		basepath, ext = os.path.splitext(self.projectfile)
		name = "jokosher-pipeline-" + os.path.basename(basepath)
		gst.DEBUG_BIN_TO_DOT_FILE_WITH_TS(self.mainpipeline, gst.DEBUG_GRAPH_SHOW_ALL, name)
		Globals.debug("Dumped pipeline to DOT file:", name)
		Globals.debug("Command to render DOT file: dot -Tsvg -o pipeline.svg <file>")
	
	#_____________________________________________________________________
	
	def SaveProjectFile(self, path=None, backup=False):
		"""
		Saves the Project and its children as an XML file
		to the path specified by file.
		
		Parameters:
			path -- path to the Project file.
		"""
		
		if not path:
			if not self.projectfile:
				raise Exception("No save path specified!")
			path = self.projectfile
		
		if not self.audio_path:
			self.audio_path = os.path.join(os.path.dirname(path), "audio")
		if not self.levels_path:
			self.levels_path = os.path.join(os.path.dirname(path), "levels")
			
		if os.path.exists(self.audio_path):
			if not os.path.isdir(self.audio_path):
				raise Exception("Audio save location is not a directory")
		else:
			os.mkdir(self.audio_path)
		
		if os.path.exists(self.levels_path):
			if not os.path.isdir(self.levels_path):
				raise Exception("Levels save location is not a directory")
		else:
			os.mkdir(self.levels_path)
		
		if not path.endswith(".jokosher"):
			path = path + ".jokosher"
			
		#sync the transport's mode with the one which will be saved
		self.transportMode = self.transport.mode
		
		if not backup:
			self.__unsavedChanges = False
			#purge main undo stack so that it will not prompt to save on exit
			self.__savedUndoStack.extend(self.__undoStack)
			self.__undoStack = []
			#purge savedRedoStack so that it will not prompt to save on exit
			self.__redoStack.extend(self.__savedRedoStack)
			self.__savedRedoStack = []
			
			# delete the incremental file since its all safe on disk now
			basepath, ext = os.path.splitext(self.projectfile)
			incr_path = basepath + self.INCREMENTAL_SAVE_EXT
			try:
				if os.path.exists(incr_path):
					os.remove(incr_path)
			except OSError:
				Globals.debug("Removal of .incremental failed! Next load we will try to restore unrestorable state!")
		
		doc = xml.Document()
		head = doc.createElement("JokosherProject")
		doc.appendChild(head)
		
		head.setAttribute("version", Globals.VERSION)
		
		params = doc.createElement("Parameters")
		head.appendChild(params)
		
		items = ["viewScale", "viewStart", "name", "author",
		         "transportMode", "bpm", "meter_nom", "meter_denom", "projectfile"]
		
		Utils.StoreParametersToXML(self, doc, params, items)
		
		notesNode = doc.createElement("Notes")
		head.appendChild(notesNode)
		
		# use repr() because XML will not preserve whitespace charaters such as \n and \t.
		notesNode.setAttribute("text", repr(self.notes))
			
		undo = doc.createElement("Undo")
		head.appendChild(undo)
		for action in self.__savedUndoStack:
			actionXML = doc.createElement("Action")
			undo.appendChild(actionXML)
			action.StoreToXML(doc, actionXML)
			
		redo = doc.createElement("Redo")
		head.appendChild(redo)
		for action in self.__redoStack:
			actionXML = doc.createElement("Action")
			redo.appendChild(actionXML)
			action.StoreToXML(doc, actionXML)
			
		for instr in self.instruments:
			instr.StoreToXML(doc, head)
			
		for instr in self.graveyard:
			instr.StoreToXML(doc, head, graveyard=True)
		
		try:
			#append "~" in case the saving fails
			gzipfile = gzip.GzipFile(path +"~", "w")
			gzipfile.write(doc.toprettyxml())
			gzipfile.close()
		except:
			os.remove(path + "~")
		else:
			#if the saving doesn't fail, move it to the proper location
			if os.path.exists(path):
				os.remove(path)
			os.rename(path + "~", path)		
		
		self.emit("undo")
	
	#_____________________________________________________________________
	
	def SaveIncrementalAction(self, action):
		if self.isDoingIncrementalRestore:
			return
		
		if self.__performingUndo or self.__performingRedo:
			return
		
		path, ext = os.path.splitext(self.projectfile)
		filename = path + self.INCREMENTAL_SAVE_EXT
		
		if self.hasDoneIncrementalSave:
			incr_file = open(filename, "a")
		else:
			# if we haven't performed an incremental save yet,
			# the existing .incremental file is old, so overwrite it.
			incr_file = open(filename, "w")
			self.hasDoneIncrementalSave = True
		
		incr_file.write(action.StoreToString())
		incr_file.write(self.INCREMENTAL_SAVE_DELIMITER)
		
		incr_file.close()
		
		self.SetUnsavedChanges()
		self.emit("incremental-save")
	
	#_____________________________________________________________________
	
	def CanDoIncrementalRestore(self):
		path, ext = os.path.splitext(self.projectfile)
		filename = path + self.INCREMENTAL_SAVE_EXT
		return os.path.exists(filename)	
	
	#_____________________________________________________________________
	
	def DoIncrementalRestore(self):
		"""
		Loads all the actions from the .incremental file and executes them
		to restore the project's state.
		"""
		
		if self.hasDoneIncrementalSave:
			Globals.debug("Cannot do incremental restore after incremental save.")
			return False
		
		path, ext = os.path.splitext(self.projectfile)
		filename = path + self.INCREMENTAL_SAVE_EXT
		
		save_action_list = []
		
		if os.path.isfile(filename):
			incr_file = open(filename, "r")
			filetext = incr_file.read()
			incr_file.close()
			for incr_xml in filetext.split(self.INCREMENTAL_SAVE_DELIMITER):
				incr_xml = incr_xml.strip()
				if incr_xml:
					incr_action = IncrementalSave.LoadFromString(incr_xml)
					save_action_list.append(incr_action)
		
		self.isDoingIncrementalRestore = True
		try:
			IncrementalSave.FilterAndExecuteAll(save_action_list, self)
		except:
			Globals.debug("Exception while restoring incremental save.",
						"Project state is surely out of sync with .incremental file")
			raise
		
		# set hasDoneIncrementSave to True because project is now in sync with .incremental file
		# i.e. we don't have to destory the .incremental file because the states match up.
		self.hasDoneIncrementalSave = True
		self.isDoingIncrementalRestore = False
		return True
		
	#_____________________________________________________________________

	def CloseProject(self):
		"""
		Closes down this Project.
		"""
		
		# when closing the file, the user chooses to either save, or discard
		# in either case, we don't need the incremental save file anymore
		path, ext = os.path.splitext(self.projectfile)
		filename = path + self.INCREMENTAL_SAVE_EXT
		try:
			if os.path.exists(filename):
				os.remove(filename)
		except OSError:
			Globals.debug("Removal of .incremental failed! Next load we will try to restore unrestorable state!")
		
		for file in self.deleteOnCloseAudioFiles:
			if os.path.exists(file):
				Globals.debug("Deleting copied audio file:", file)
				os.remove(file)
		self.deleteOnCloseAudioFiles = []
		
		self.mainpipeline.set_state(gst.STATE_NULL)
		
	#_____________________________________________________________________
	
	def Undo(self):
		"""
		Attempts to revert the last user action by popping an action
		from the undo stack and executing it.
		"""
		self.__performingUndo = True
		
		if len(self.__undoStack):
			cmd = self.__undoStack.pop()
			self.ExecuteAction(cmd)
			
		elif len(self.__savedUndoStack):
			self.__savedUndo = True
			cmd = self.__savedUndoStack.pop()
			self.ExecuteAction(cmd)
			self.__savedUndo = False
			
		self.__performingUndo = False
		
		# __performingUndo must be False for project to log
		inc = IncrementalSave.Undo()
		self.SaveIncrementalAction(inc)
	
	#_____________________________________________________________________
	
	def Redo(self):
		"""
		Attempts to redo the last undone action.
		"""
		self.__performingRedo = True
		
		if len(self.__savedRedoStack):
			self.__savedUndo = True
			cmd = self.__savedRedoStack.pop()
			self.ExecuteAction(cmd)
			self.__savedUndo = False
			
		elif len(self.__redoStack):
			cmd = self.__redoStack.pop()
			self.ExecuteAction(cmd)
			
		self.__performingRedo = False
		
		# __performingRedo must be False for project to log
		inc = IncrementalSave.Redo()
		self.SaveIncrementalAction(inc)

	#_____________________________________________________________________
	
	def AppendToCurrentStack(self, object):
		"""
		Appends the action specified by object onto the relevant
		undo/redo stack.
		
		Parameters:
			object -- action to be added to the undo/redo stack
		"""
		if self.__savedUndo and self.__performingUndo:
			self.__savedRedoStack.append(object)
		elif self.__savedUndo and self.__performingRedo:
			self.__savedUndoStack.append(object)
		elif self.__performingUndo:
			self.__redoStack.append(object)
		elif self.__performingRedo:
			self.__undoStack.append(object)
		else:
			self.__undoStack.append(object)
			self.__redoStack = []
			#if we have undone anything that was previously saved
			if len(self.__savedRedoStack):
				self.__savedRedoStack = []
				#since there is no other record that something has 
				#changed after savedRedoStack is purged
				self.__unsavedChanges = True
		self.emit("undo")
	
	#_____________________________________________________________________
	
	def NewAtomicUndoAction(self):
		"""
		Creates a new AtomicUndoAction and adds to the
		appropriate undo/redo stack for this project.
		
		Return:
			The newly created AtomicUndoAction instance.
		"""
		undoAction = UndoSystem.AtomicUndoAction()
		self.AppendToCurrentStack(undoAction)
		return undoAction
		
	#_____________________________________________________________________
	
	def CheckUnsavedChanges(self):
		"""
		Uses boolean self.__unsavedChanges and Undo/Redo to 
		determine if the program needs to save anything on exit.
		
		Return:
			True -- there's unsaved changes, undoes or redoes
			False -- the Project can be safely closed.
		"""
		return self.__unsavedChanges or \
			len(self.__undoStack) > 0 or \
			len(self.__savedRedoStack) > 0
	
	#_____________________________________________________________________
	
	def SetUnsavedChanges(self):
		self.__unsavedChanges = True
		self.emit("undo") 
		
	#_____________________________________________________________________
	
	def CanPerformUndo(self):
		"""
		Whether it's possible to perform an undo operation.
		
		Returns:
			True -- there is another undo command in the stack that can be performed.
			False -- there are no available undo commands.
		"""
		return bool(len(self.__undoStack) or len(self.__savedUndoStack))
	
	#_____________________________________________________________________
	
	def CanPerformRedo(self):
		"""
		Whether it's possible to perform an redo operation.
		
		Returns:
			True -- there is another redo command in the stack that can be performed.
			False -- there are no available redo commands.
		"""
		return bool(len(self.__redoStack) or len(self.__savedRedoStack))
	
	#_____________________________________________________________________
	
	def ExecuteAction(self, undoAction):
		"""
		Executes an AtomicUndoAction object, reverting all operations stored
		in it.
			
		Parameters:
			undoAction -- the instance of AtomicUndoAction to be executed.
		"""
		newUndoAction = self.NewAtomicUndoAction()
		for cmdList in reversed(undoAction.GetUndoCommands()):
			obj = cmdList[0]
			target_object = self.JokosherObjectFromString(obj)
			
			getattr(target_object, cmdList[1])(_undoAction_=newUndoAction, *cmdList[2])

	#_____________________________________________________________________
	
	def JokosherObjectFromString(self, string):
		"""
		Converts a string used to serialize references to Project, Instrument
		and Event instances into a reference to the actual object.
		
		Parameters:
			string -- The string to convert such as "P" for project or "I2" for instrument with ID equal to 2.
		"""
		if string[0] == "P":		# Check if the object is a Project
			return self
		elif string[0] == "I":		# Check if the object is an Instrument
			id = int(string[1:])
			for instr in self.instruments:
				if instr.id == id:
					return instr
		elif string[0] == "E":		# Check if the object is an Event
			id = int(string[1:])
			for instr in self.instruments:
				for event in instr.events:
					if event.id == id:
						return event
				for event in instr.graveyard:
					if event.id == id:
						return event
				
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("SetBPM", "temp")
	def SetBPM(self, bpm):
		"""
		Changes the current beats per minute.
		
		Parameters:
			bpm -- value of the new beats per minute.
		"""
		self.temp = self.bpm
		if self.bpm != bpm:
			self.bpm = bpm
			#FIXME: find a better way to do project.PrepareClick() it doesn't take a really long time with large bpm
			self.PrepareClick()
			self.emit("bpm")
	
	#_____________________________________________________________________

	@UndoSystem.UndoCommand("SetMeter", "temp", "temp1")
	def SetMeter(self, nom, denom):
		"""
		Changes the current time signature.
		
		Example:
			nom = 3
			denom = 4
			
			would result in the following signature:
				3/4
		
		Parameters:
			nom -- new time signature nominator.
			denom --new time signature denominator.
		"""
		self.temp = self.meter_nom
		self.temp1 = self.meter_denom
		
		if self.meter_nom != nom or self.meter_denom != denom:
			self.meter_nom = nom
			self.meter_denom = denom
			self.emit("time-signature")
			
	#_____________________________________________________________________
	
	def AddInstruments(self, instrTuples):
		"""
		Adds one or more instruments to the Project, and ensures that
		they are all appended to the undo stack as a single atomic action.
		
		Parameters:
			instrTuples -- a list of tuples containing name and type
					that will be passed to AddInstrument().
			
		Returns:
			A list of the added Instruments.
		"""
		
		undoAction = self.NewAtomicUndoAction()
		instrList = []
		for name, type in instrTuples:
			instr = self.AddInstrument(name, type, _undoAction_=undoAction)
			instrList.append(instr)
		return instrList
	
	#_____________________________________________________________________
	
	def DeleteInstrumentsOrEvents(self, instrumentOrEventList):
		"""
		Removes the given instruments the Project.
		
		Parameters:
			instrumentList -- a list of Instrument instances to be removed.
		"""
		undoAction = self.NewAtomicUndoAction()
		for instrOrEvent in instrumentOrEventList:
			if isinstance(instrOrEvent, Instrument.Instrument):
				self.DeleteInstrument(instrOrEvent.id, _undoAction_=undoAction)
			elif isinstance(instrOrEvent, Event.Event):
				instrOrEvent.instrument.DeleteEvent(instrOrEvent.id, _undoAction_=undoAction)
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("DeleteInstrument", "temp")
	def AddInstrument(self, name, type):
		"""
		Adds a new instrument to the Project and returns the ID for that instrument.
		
		Considerations:
			In most cases, AddInstruments() should be used instead of this function
			to ensure that the undo actions are made atomic.
		
		Parameters:
			name -- name of the instrument.
			type -- type of the instrument.
			
		Returns:
			The created Instrument object.
		"""
		pixbuf = Globals.getCachedInstrumentPixbuf(type)
		instr = Instrument.Instrument(self, name, type, pixbuf)
		if len(self.instruments) == 0:
			#If this is the first instrument, arm it by default
			instr.isArmed = True
		
		self.temp = instr.id
		self.instruments.append(instr)
		
		self.emit("instrument::added", instr)
		return instr
		
	#_____________________________________________________________________	
	
	@UndoSystem.UndoCommand("ResurrectInstrument", "temp")
	def DeleteInstrument(self, id):
		"""
		Removes the instrument matching id from the Project.
		
		Considerations:
			In most cases, DeleteInstrumentsOrEvents() should be used instead
			of this function to ensure that the undo actions are made atomic.
		
		Parameters:
			id -- unique ID of the instument to remove.
		"""
		
		instrs = [x for x in self.instruments if x.id == id]
		if not instrs:
			raise UndoSystem.CancelUndoCommand()
		
		instr = instrs[0]
		instr.RemoveAndUnlinkPlaybackbin()
		
		self.graveyard.append(instr)
		self.instruments.remove(instr)
		if instr.isSolo:
			self.soloInstrCount -= 1
			self.OnAllInstrumentsMute()
		
		for event in instr.events:
			event.StopGenerateWaveform(False)
			
		self.temp = id
		self.emit("instrument::removed", instr)
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("DeleteInstrument", "temp")
	def ResurrectInstrument(self, id):
		"""
		Brings a deleted Instrument back from the graveyard.
		
		Parameters:
			id -- unique ID of the instument to restore.
		"""
		instr = [x for x in self.graveyard if x.id == id][0]
		
		instr.AddAndLinkPlaybackbin()
		
		self.instruments.append(instr)
		if instr.isSolo:
			self.soloInstrCount += 1
			self.OnAllInstrumentsMute()
			
		for event in instr.events:
			if event.isLoading:
				event.GenerateWaveform()
		
		instr.isVisible = True
		self.graveyard.remove(instr)
		self.temp = id
		self.emit("instrument::added", instr)
		
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("MoveInstrument", "temp", "temp1")
	def MoveInstrument(self, id, position):
		"""
		Move an instrument in the instrument list.
		Used for drag and drop ordering of instruments in InstrumentViewer.py
		
		Parameters:
			id -- unique ID of the instument to restore.
			position -- new position of the instrument inside the instrument 
						pane to the left of the screen.
		"""
		self.temp = id
		instr = [x for x in self.instruments if x.id == id][0]
		self.temp1 = self.instruments.index(instr)
		
		self.instruments.remove(instr)
		self.instruments.insert(position, instr)
		self.emit("instrument::reordered", instr)
	
	#_____________________________________________________________________
	
	def AddInstrumentAndEvents(self, fileList, copyFile=False, undoAction=None):
		"""
		Adds a new instrument of type "audio file" to the project, and adds the
		given files to the new instrument.
		
		Parameters:
			fileList -- the list of paths of the files that should be added.
			copyFile -- True = copy the files to Project's audio directory.
					False = don't copy the files to the Project's audio directory.
		"""
		if not undoAction:
			undoAction = self.NewAtomicUndoAction()
	
		uris = [PlatformUtils.pathname2url(filename) for filename in fileList]

		name, type, pixbuf, path = [x for x in Globals.getCachedInstruments() if x[1] == "audiofile"][0]
		instr = self.AddInstrument(name, type, _undoAction_=undoAction)
		instr.AddEventsFromList(0, uris, copyFile, undoAction)
	
	#_____________________________________________________________________
	
	def ClearEventSelections(self):
		"""
		Clears the selection of any events.
		"""
		for instr in self.instruments:
			for event in instr.events:
				event.SetSelected(False)

	#_____________________________________________________________________

	def SelectInstrument(self, instrument=None):
		"""
		Selects an instrument and clears the selection of all other instruments.
		
		Parameters:
			instrument -- Instrument object corresponding to the selected instrument.
		"""
		for instr in self.instruments:
			if instr is not instrument:
				instr.SetSelected(False)
			else:
				instr.SetSelected(True)
			
	#_____________________________________________________________________
	
	def SetViewStart(self, start):
		"""
		Sets the time at which the Project view should start.

		Parameters:
			start -- start time for the view in seconds.
		"""
		start = max(0, min(self.GetProjectLength(), start))
		if self.viewStart != start:
			self.viewStart = start
			self.emit("view-start")
		
	#_____________________________________________________________________
	
	def SetViewScale(self, scale):
		"""
		Sets the scale of the Project view.
		
		Parameters:
			scale -- view scale in pixels per second.
		"""
		self.viewScale = scale
		self.emit("zoom")
		
	#_____________________________________________________________________

	def GetProjectLength(self):
		"""
		Returns the length of the Project.
		
		Returns:
			lenght of the Project in seconds.
		"""
		length = 0
		for instr in self.instruments:
			for event in instr.events:
				size = event.start + max(event.duration, event.loadingLength)
				length = max(length, size)
		return length

	#_____________________________________________________________________
	
	def OnAllInstrumentsMute(self):
		"""
		Mutes all Instruments in this Project.
		"""
		for instr in self.instruments:
			instr.OnMute()
			
	#_____________________________________________________________________
	
	def GenerateUniqueID(self, id = None,  reserve=True):
		"""
		Creates a new unique ID which can be assigned to an new Project object.
		
		Parameters:
			id -- an unique ID proposal. If it's already taken, a new one is generated.
			reserve -- if True, the ID will be recorded and never returned again.
			
		Returns:
			an unique ID suitable for a new Project.
		"""
		if id != None:
			if id in self.___id_list:
				Globals.debug("Error: id", id, "already taken")
			else:
				if reserve:
					self.___id_list.append(id)
				return id
				
		counter = 0
		while True:
			if not counter in self.___id_list:
				if reserve:
					self.___id_list.append(counter)
				return counter
			counter += 1
	
	#_____________________________________________________________________

	def SetVolume(self, volume):
		"""
		Sets the volume of an instrument.
		
		Parameters:
			volume - a value in the range [0,1]
		"""
		self.volume = volume
		for instr in self.instruments:
			instr.UpdateVolume()
		self.emit("volume")

	#_____________________________________________________________________

	def SetLevel(self, level):
		"""
		Sets the current REPORTED level, NOT THE VOLUME!
		
		Parameters:
			level -- a value in the range [0,1]
		"""
		self.level = level

	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("SetTransportMode", "temp")
	def SetTransportMode(self, val):
		"""
		Sets the Mode in the Transportmanager. Used to enable Undo/Redo.
		
		Parameters:
			val -- the mode to display the timeline bar:
					TransportManager.MODE_HOURS_MINS_SECS
					TransportManager.MODE_BARS_BEATS
		"""
		self.temp = self.transport.mode
		self.transport.SetMode(val)

	#_____________________________________________________________________

	def PrepareClick(self):
		"""
		Prepares the click track.
		"""

		self.ClearClickTimes()

		second = 1000000000
		
		# FIXME: currently hard coded to 600 seconds
		length = (600 * second)
		interval = second / (self.bpm/60)
		
		self.clickTrackController.set("volume", 0 * gst.SECOND, 0.0)

		current = 0 + interval

		while current < length:
			self.clickTrackController.set("volume", current-(second / 10), 0.0)
			self.clickTrackController.set("volume", current, 1.0)
			self.clickTrackController.set("volume", current+(second / 10), 0.0)

			current = current + interval

	#_____________________________________________________________________

	def SetClickTrackVolume(self, value):
		"""
		Unmutes and enables the click track.
		
		Parameters:
			value -- The volume of the click track between 0.0 and 1.0
		"""
		if self.clickVolumeValue != value:
			self.clickTrackVolume.set_property("mute", (value < 0.01))
			# convert the 0.0 to 1.0 range to 0.0 to 2.0 range (to let the user make it twice as loud)
			self.clickTrackVolume.set_property("volume", value * 2)
			self.clickVolumeValue = value
			self.emit("click-track", value)

	#_____________________________________________________________________

	def ClearClickTimes(self):
		"""
		Clears the click track controller times.
		"""
		self.clickTrackController.unset_all("volume")
		
	#_____________________________________________________________________
	
	def SetProjectSink(self):
		"""
		Grabs the sink element based on the Global preferences, and sets
		the pipeline to use that sink.
		"""
		
		if self.audioState == self.AUDIO_EXPORTING:
			#we're exporting so some encoders and a filesink are hooked up
			#changing that would mess everything up.
			return
		
		if self.audioState != self.AUDIO_STOPPED:
			self.Stop()
		
		self.levelElement.unlink(self.masterSink)
		self.playbackbin.remove(self.masterSink)
		self.masterSink.set_state(gst.STATE_NULL)
		
		self.masterSink = self.MakeProjectSink()
		self.playbackbin.add(self.masterSink)
		self.levelElement.link(self.masterSink)
	
	#_____________________________________________________________________
	
	def SetProjectSinkDevice(self):
		"""
		Grabs the sink element device based on the Global preferences, and sets
		the pipeline to use that device.
		"""
		if self.audioState != self.AUDIO_EXPORTING and \
		   self.audioState != self.AUDIO_STOPPED:
			self.Stop()
			
		if not self.masterSink:
			return
		
		sinkElement = self.masterSink.sinks().next()
		if hasattr(sinkElement.props, "device"):
			outdevice = Globals.settings.playback["device"]
			Globals.debug("Changing output device: %s" % outdevice)
			sinkElement.set_property("device", outdevice)
		
	#_____________________________________________________________________
	
	def MakeProjectSink(self):
		"""
		Contructs a GStreamer sink element (or bin with ghost pads) for the 
		Project's audio output, according to the Global "audiosink" property.
		
		Return:
			the newly created GStreamer sink element.
		"""
		
		sinkString = Globals.settings.playback["audiosink"]
		if self.currentSinkString == sinkString:
			return self.masterSink
		
		self.currentSinkString = sinkString
		sinkBin = None
		
		try:
			sinkBin = gst.parse_bin_from_description(sinkString, True)
		except gobject.GError:
			Globals.debug("Parsing failed: %s" % sinkString)
			# autoaudiosink is our last resort
			sinkBin = gst.element_factory_make("autoaudiosink")
			Globals.debug("Using autoaudiosink for audio output")
		else:
			Globals.debug("Using custom pipeline for audio sink: %s" % sinkString)
			
			sinkElement = sinkBin.sinks().next()
			if hasattr(sinkElement.props, "device"):
				outdevice = Globals.settings.playback["device"]
				Globals.debug("Output device: %s" % outdevice)
				sinkElement.set_property("device", outdevice)
		
		return sinkBin
		
	#____________________________________________________________________	
	
	def OnCaptureBackendChange(self):
		for instr in self.instruments:
			instr.input = None
			instr.inTrack = -1
	
	#____________________________________________________________________	
	
	def GetInputFilenames(self):
		"""
		Obtains a list of  all filenames that are to be input to
		the pipeline.
		
		Return: 
			a list of  the filenames
		"""
		fileList = []
		for instrument in self.instruments:
			for event in instrument.events:
				fileList.append(event.GetAbsFile())
		return fileList
		
	#____________________________________________________________________	
	
	def GetLocalAudioFilenames(self):
		fileList = []
		for instrument in self.instruments:
			for event in instrument.events:
				if not os.path.isabs(event.file):
					fileList.append(event.file)
		return fileList
	
	#____________________________________________________________________	
	
	def GetLevelsFilenames(self):
		fileList = []
		for instrument in self.instruments:
			for event in instrument.events:
				fileList.append(event.levels_file)
		return fileList
	
	#____________________________________________________________________	
	

	def SetName(self, name):
		if self.name != name:
			self.name = name
			inc = IncrementalSave.SetName(name)
			self.SaveIncrementalAction(inc)
				
	#____________________________________________________________________	
	
	def SetAuthor(self, author):
		if self.author != author:
			self.author = author
			inc = IncrementalSave.SetAuthor(author)
			self.SaveIncrementalAction(inc)
	
	#____________________________________________________________________	
	
	def SetNotes(self, notes):
		if self.notes != notes:
			self.notes = notes
			inc = IncrementalSave.SetNotes(notes)
			self.SaveIncrementalAction(inc)
	
	#____________________________________________________________________		
#=========================================================================
